# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import random
import string
import time

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.modules.util import finalize
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.protocol import Node
from nbxmpp.protocol import NodeProcessed
from nbxmpp.protocol import StanzaMalformed
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import PGPKeyMetadata
from nbxmpp.structs import PGPPublicKey
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task
from nbxmpp.util import b64decode
from nbxmpp.util import b64encode

if TYPE_CHECKING:
    from nbxmpp.client import Client


class OpenPGP(BaseModule):

    _depends = {
        "publish": "PubSub",
        "request_items": "PubSub",
    }

    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_pubsub_openpgp,
                ns=Namespace.PUBSUB_EVENT,
                priority=16,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_openpgp_message,
                ns=Namespace.OPENPGP,
                priority=7,
            ),
        ]

    def _process_openpgp_message(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        openpgp = stanza.getTag("openpgp", namespace=Namespace.OPENPGP)
        if openpgp is None:
            self._log.warning("No openpgp node found")
            self._log.warning(stanza)
            return

        data = openpgp.getData()
        if not data:
            self._log.warning("No data in openpgp node found")
            self._log.warning(stanza)
            return

        self._log.info("Encrypted message received")
        try:
            properties.openpgp = b64decode(data)
        except Exception:
            self._log.warning("b64decode failed")
            self._log.warning(stanza)
            return

    def _process_pubsub_openpgp(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        """
        <item>
            <public-keys-list xmlns='urn:xmpp:openpgp:0'>
              <pubkey-metadata
                v4-fingerprint='1357B01865B2503C18453D208CAC2A9678548E35'
                date='2018-03-01T15:26:12Z'
                />
              <pubkey-metadata
                v4-fingerprint='67819B343B2AB70DED9320872C6464AF2A8E4C02'
                date='1953-05-16T12:00:00Z'
                />
            </public-keys-list>
        </item>
        """

        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.OPENPGP_PK:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        try:
            data = _parse_keylist(properties.jid, item)
        except ValueError as error:
            self._log.warning(error)
            self._log.warning(stanza)
            raise NodeProcessed

        if data is None:
            self._log.info("Received PGP keylist: %s - no keys set", properties.jid)
            return

        pubsub_event = properties.pubsub_event._replace(data=data)
        self._log.info("Received PGP keylist: %s - %s", properties.jid, data)

        properties.pubsub_event = pubsub_event

    @iq_request_task
    def set_keylist(
        self, keylist: list[PGPKeyMetadata] | None, public: bool = True
    ) -> Any:
        task = yield

        access_model = "open" if public else "presence"

        options = {
            "pubsub#persist_items": "true",
            "pubsub#access_model": access_model,
        }

        result = yield self.publish(
            Namespace.OPENPGP_PK,
            _make_keylist(keylist),
            id_="current",
            options=options,
            force_node_options=True,
        )

        yield finalize(task, result)

    @iq_request_task
    def set_public_key(
        self, key: bytes, fingerprint: str, date: float, public: bool = True
    ) -> Any:
        task = yield

        access_model = "open" if public else "presence"

        options = {
            "pubsub#persist_items": "true",
            "pubsub#access_model": access_model,
        }

        result = yield self.publish(
            f"{Namespace.OPENPGP_PK}:{fingerprint}",
            _make_public_key(key, date),
            id_="current",
            options=options,
            force_node_options=True,
        )

        yield finalize(task, result)

    @iq_request_task
    def request_public_key(self, jid: JID, fingerprint: str) -> Any:
        task = yield

        items = yield self.request_items(
            f"{Namespace.OPENPGP_PK}:{fingerprint}", max_items=1, jid=jid
        )

        raise_if_error(items)

        if not items:
            yield task.set_result(None)

        try:
            key = _parse_public_key(jid, items[0])
        except ValueError as error:
            raise MalformedStanzaError(str(error), items)

        yield key

    @iq_request_task
    def request_keylist(self, jid: JID | None = None) -> Any:
        task = yield

        items = yield self.request_items(Namespace.OPENPGP_PK, max_items=1, jid=jid)

        raise_if_error(items)

        if not items:
            yield task.set_result(None)

        try:
            keylist = _parse_keylist(jid, items[0])
        except ValueError as error:
            raise MalformedStanzaError(str(error), items)

        self._log.info("Received keylist: %s", keylist)
        yield keylist

    @iq_request_task
    def request_secret_key(self) -> Any:
        task = yield

        items = yield self.request_items(Namespace.OPENPGP_SK, max_items=1)

        raise_if_error(items)

        if not items:
            yield task.set_result(None)

        try:
            secret_key = _parse_secret_key(items[0])
        except ValueError as error:
            raise MalformedStanzaError(str(error), items)

        yield secret_key

    @iq_request_task
    def set_secret_key(self, secret_key: bytes) -> Any:
        task = yield

        options = {
            "pubsub#persist_items": "true",
            "pubsub#access_model": "whitelist",
        }

        self._log.info("Set secret key")

        result = yield self.publish(
            Namespace.OPENPGP_SK,
            _make_secret_key(secret_key),
            id_="current",
            options=options,
            force_node_options=True,
        )

        yield finalize(task, result)


def parse_signcrypt(stanza: Node) -> tuple[list[Node | str], list[JID], str]:
    """
    <signcrypt xmlns='urn:xmpp:openpgp:0'>
      <to jid='juliet@example.org'/>
      <time stamp='2014-07-10T17:06:00+02:00'/>
      <rpad>
        f0rm1l4n4-mT8y33j!Y%fRSrcd^ZE4Q7VDt1L%WEgR!kv
      </rpad>
      <payload>
        <body xmlns='jabber:client'>
          This is a secret message.
        </body>
      </payload>
    </signcrypt>
    """
    if stanza.getName() != "signcrypt" or stanza.getNamespace() != Namespace.OPENPGP:
        raise StanzaMalformed("Invalid signcrypt node")

    to_nodes = stanza.getTags("to")
    if not to_nodes:
        raise StanzaMalformed("missing to nodes")

    recipients: list[JID] = []
    for to_node in to_nodes:
        jid = to_node.getAttr("jid")
        try:
            recipients.append(JID.from_string(jid))
        except Exception as error:
            raise StanzaMalformed("Invalid jid: %s %s" % (jid, error))

    timestamp = stanza.getTagAttr("time", "stamp")
    if timestamp is None:
        raise StanzaMalformed("Invalid timestamp")

    payload = stanza.getTag("payload")
    if payload is None or not payload.getChildren():
        raise StanzaMalformed("Invalid payload node")
    return payload.getChildren(), recipients, timestamp


def create_signcrypt_node(
    stanza: Node, recipients: list[JID], not_encrypted_nodes: list[tuple[str, str]]
) -> Node:
    """
    <signcrypt xmlns='urn:xmpp:openpgp:0'>
      <to jid='juliet@example.org'/>
      <time stamp='2014-07-10T17:06:00+02:00'/>
      <rpad>
        f0rm1l4n4-mT8y33j!Y%fRSrcd^ZE4Q7VDt1L%WEgR!kv
      </rpad>
      <payload>
        <body xmlns='jabber:client'>
          This is a secret message.
        </body>
      </payload>
    </signcrypt>
    """
    encrypted_nodes: list[Node] = []
    child_nodes = list(stanza.getChildren())
    for node in child_nodes:
        if isinstance(node, str):
            stanza.delChild(node)
            continue

        if (node.getName(), node.getNamespace()) not in not_encrypted_nodes:
            if not node.getNamespace():
                node.setNamespace(Namespace.CLIENT)
            encrypted_nodes.append(node)
            stanza.delChild(node)

    signcrypt = Node("signcrypt", attrs={"xmlns": Namespace.OPENPGP})
    for recipient in recipients:
        signcrypt.addChild("to", attrs={"jid": str(recipient)})

    timestamp = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
    signcrypt.addChild("time", attrs={"stamp": timestamp})

    signcrypt.addChild("rpad").addData(get_rpad())

    payload = signcrypt.addChild("payload")

    for node in encrypted_nodes:
        payload.addChild(node=node)

    return signcrypt


def get_rpad() -> str:
    rpad_range = random.randint(30, 50)
    return "".join(random.choice(string.ascii_letters) for _ in range(rpad_range))


def create_message_stanza(
    stanza: Message, encrypted_payload: str | bytes, with_fallback_text: bool
) -> None:
    b64encoded_payload = b64encode(encrypted_payload)

    openpgp_node = Node("openpgp", attrs={"xmlns": Namespace.OPENPGP})
    openpgp_node.addData(b64encoded_payload)
    stanza.addChild(node=openpgp_node)

    eme_node = Node(
        "encryption", attrs={"xmlns": Namespace.EME, "namespace": Namespace.OPENPGP}
    )
    stanza.addChild(node=eme_node)

    if with_fallback_text:
        stanza.setBody("This message is *encrypted* with OpenPGP")


def _make_keylist(keylist: list[PGPKeyMetadata] | None) -> Node:
    item = Node("public-keys-list", {"xmlns": Namespace.OPENPGP})
    if keylist is not None:
        for key in keylist:
            date = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(key.date))
            attrs = {"v4-fingerprint": key.fingerprint, "date": date}
            item.addChild("pubkey-metadata", attrs=attrs)
    return item


def _make_public_key(key: str | bytes, date: float) -> Node:
    date_str = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(date))
    item = Node("pubkey", attrs={"xmlns": Namespace.OPENPGP, "date": date_str})
    data = item.addChild("data")
    data.addData(b64encode(key))
    return item


def _make_secret_key(secret_key: str | bytes | None) -> Node:
    item = Node("secretkey", {"xmlns": Namespace.OPENPGP})
    if secret_key is not None:
        item.setData(b64encode(secret_key))
    return item


def _parse_public_key(jid: JID, item: Node) -> PGPPublicKey:
    pub_key = item.getTag("pubkey", namespace=Namespace.OPENPGP)
    if pub_key is None:
        raise ValueError("pubkey node missing")

    date = parse_datetime(pub_key.getAttr("date"), epoch=True)
    assert isinstance(date, float)

    data = pub_key.getTag("data")
    if data is None:
        raise ValueError("data node missing")

    try:
        key = b64decode(data.getData())
    except Exception as error:
        raise ValueError(f"decoding error: {error}")

    return PGPPublicKey(jid, key, date)


def _parse_keylist(jid: JID, item: Node) -> list[PGPKeyMetadata] | None:
    keylist_node = item.getTag("public-keys-list", namespace=Namespace.OPENPGP)
    if keylist_node is None:
        raise ValueError("public-keys-list node missing")

    metadata = keylist_node.getTags("pubkey-metadata")
    if not metadata:
        return None

    data: list[PGPKeyMetadata] = []
    for key in metadata:
        fingerprint = key.getAttr("v4-fingerprint")
        date = key.getAttr("date")
        if fingerprint is None or date is None:
            raise ValueError("Invalid metadata node")

        timestamp = parse_datetime(date, epoch=True)
        if not isinstance(timestamp, float):
            raise ValueError("Invalid date timestamp: %s" % date)

        data.append(PGPKeyMetadata(jid, fingerprint, timestamp))
    return data


def _parse_secret_key(item: Node) -> bytes:
    sec_key = item.getTag("secretkey", namespace=Namespace.OPENPGP)
    if sec_key is None:
        raise ValueError("secretkey node missing")

    data = sec_key.getData()
    if not data:
        raise ValueError("secretkey data missing")

    try:
        key = b64decode(data)
    except Exception as error:
        raise ValueError(f"decoding error: {error}")

    return key



# Copyright (C) 2022 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import Reactions as ReactionStruct
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client
    from nbxmpp.protocol import Message


class Reactions(BaseModule):
    def __init__(self, client: Client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message_reaction,
                ns=Namespace.REACTIONS,
                priority=15,
            ),
        ]

    def _process_message_reaction(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        reactions = stanza.getTag("reactions", namespace=Namespace.REACTIONS)

        if reactions is None:
            return

        id_ = reactions.getAttr("id")
        if not id_:
            self._log.warning("Reactions without ID")
            return

        emojis: set[str] = set()
        for reaction in reactions.getTags("reaction"):
            # we strip for clients that might add white spaces and/or
            # new lines in the reaction content.
            emoji = reaction.getData().strip()
            if emoji:
                emojis.add(emoji)
            else:
                self._log.warning("Empty reaction")
                self._log.warning(stanza)

        properties.reactions = ReactionStruct(id_, emojis)



def xml2str_sorted(data):
    s = "<" + data.name
    if data.namespace:
        if not data.parent or data.parent.namespace != data.namespace:
            if "xmlns" not in data.attrs:
                s += ' xmlns="%s"' % data.namespace
    for key in sorted(data.attrs.keys()):
        val = str(data.attrs[key])
        s += ' %s="%s"' % (key, val)

    s += ">"
    cnt = 0
    if data.kids:
        for a in data.kids:
            if (len(data.data) - 1) >= cnt:
                s += data.data[cnt]
            if isinstance(a, str):
                s += a.__str__()
            else:
                s += xml2str_sorted(a)
            cnt += 1
    if (len(data.data) - 1) >= cnt:
        s += data.data[cnt]
    if not data.kids and s.endswith(">"):
        s = s[:-1] + " />"
    else:
        s += "</" + data.name + ">"
    return s



import datetime
from test.lib.util import StanzaHandlerTest

from nbxmpp import Message
from nbxmpp.modules.muc.moderation import Moderation
from nbxmpp.protocol import JID
from nbxmpp.structs import MessageProperties


class TestModeration(StanzaHandlerTest):
    def test_parse_moderation_message(self):
        xml = """
        <message type="groupchat" id='retraction-id-1' from='room@muc.example.com' to="room@muc.example.com/macbeth">
          <retract id="stanza-id-1" xmlns='urn:xmpp:message-retract:1'>
            <moderated by='room@muc.example.com/macbeth' xmlns='urn:xmpp:message-moderate:1'>
              <occupant-id xmlns="urn:xmpp:occupant-id:0" id="dd72603deec90a38ba552f7c68cbcc61bca202cd" />
            </moderated>
            <reason>Some Reason</reason>
          </retract>
        </message>
        """

        props = MessageProperties("mockjid")
        message = Message(node=xml)
        Moderation._process_moderation_1_message(None, None, message, props)

        assert props.moderation is not None
        self.assertEqual(props.moderation.stanza_id, "stanza-id-1")
        self.assertEqual(
            props.moderation.by, JID.from_string("room@muc.example.com/macbeth")
        )
        self.assertEqual(props.moderation.reason, "Some Reason")
        self.assertIsInstance(props.moderation.stamp, datetime.datetime)
        self.assertFalse(props.moderation.is_tombstone)
        self.assertEqual(
            props.moderation.occupant_id, "dd72603deec90a38ba552f7c68cbcc61bca202cd"
        )



import sys
import unittest
from test.lib.const import STREAM_START
from unittest.mock import Mock

from nbxmpp.dispatcher import StanzaDispatcher
from nbxmpp.protocol import JID


class StanzaHandlerTest(unittest.TestCase):
    def setUp(self):
        # Setup mock client
        self.client = Mock()
        self.client.is_websocket = False
        self.dispatcher = StanzaDispatcher(self.client)

        self.client.get_bound_jid.return_value = JID.from_string("test@test.test")

        self.dispatcher.reset_parser()
        self.dispatcher.process_data(STREAM_START)


def raise_all_exceptions(func):
    # Exceptions which are raised from async callbacks
    # in GLib or GTK do not bubble up to the unittest
    # This decorator catches all exceptions and raises them
    # after the unittest
    def func_wrapper(self, *args, **kwargs):

        exceptions = []

        def on_hook(type_, value, tback):
            exceptions.append((type_, value, tback))

        orig_excepthook = sys.excepthook
        sys.excepthook = on_hook
        try:
            result = func(self)
        finally:
            sys.excepthook = orig_excepthook
            if exceptions:
                tp, value, tb = exceptions[0]
                raise tp(value).with_traceback(tb)

        return result

    return func_wrapper



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

import json
from collections.abc import Generator

from nbxmpp.const import AnonymityMode
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.http import HTTPRequest
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.dataforms import extend_form
from nbxmpp.modules.util import finalize
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import Node
from nbxmpp.structs import MuclumbusItem
from nbxmpp.structs import MuclumbusResult
from nbxmpp.task import http_request_task
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client

# API Documentation
# https://search.jabber.network/docs/api


class Muclumbus(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def request_parameters(self, jid: str):
        task = yield

        response = yield _make_parameter_request(jid)
        if response.isError():
            raise StanzaError(response)

        search = response.getTag("search", namespace=Namespace.MUCLUMBUS)
        if search is None:
            raise MalformedStanzaError("search node missing", response)

        dataform = search.getTag("x", namespace=Namespace.DATA)
        if dataform is None:
            raise MalformedStanzaError("dataform node missing", response)

        self._log.info("Muclumbus parameters received")
        yield finalize(task, extend_form(node=dataform))

    @iq_request_task
    def set_search(
        self, jid: str, dataform, items_per_page: int = 50, after: str | None = None
    ):
        _task = yield

        response = yield _make_search_query(jid, dataform, items_per_page, after)
        if response.isError():
            raise StanzaError(response)

        result = response.getTag("result", namespace=Namespace.MUCLUMBUS)
        if result is None:
            raise MalformedStanzaError("result node missing", response)

        items = result.getTags("item")
        if not items:
            yield MuclumbusResult(first=None, last=None, max=None, end=True, items=[])

        set_ = result.getTag("set", namespace=Namespace.RSM)
        if set_ is None:
            raise MalformedStanzaError("set node missing", response)

        first = set_.getTagData("first")
        last = set_.getTagData("last")
        try:
            max_ = int(set_.getTagData("max"))
        except Exception:
            raise MalformedStanzaError("invalid max value", response)

        results: list[MuclumbusItem] = []
        for item in items:
            jid = item.getAttr("address")
            name = item.getTagData("name")
            nusers = item.getTagData("nusers")
            description = item.getTagData("description")
            language = item.getTagData("language")
            is_open = item.getTag("is-open") is not None

            try:
                anonymity_mode = AnonymityMode(item.getTagData("anonymity-mode"))
            except ValueError:
                anonymity_mode = AnonymityMode.UNKNOWN
            results.append(
                MuclumbusItem(
                    jid=jid,
                    name=name or "",
                    nusers=nusers or "",
                    description=description or "",
                    language=language or "",
                    is_open=is_open,
                    anonymity_mode=anonymity_mode,
                )
            )
        yield MuclumbusResult(
            first=first, last=last, max=max_, end=len(items) < max_, items=results
        )

    @http_request_task
    def set_http_search(
        self, uri: str, keywords: list[str], after: str | None = None
    ) -> Generator[MuclumbusResult, HTTPRequest, None]:
        _task = yield

        search = {"keywords": keywords}
        if after is not None:
            search["after"] = after

        body = json.dumps(search).encode()

        session = self._client.http_session
        request = session.create_request()
        request.set_request_body("application/json", body)
        request.send("POST", uri)

        request = yield request

        if not request.is_complete():
            self._log.warning(request.get_error_string())
            yield MuclumbusResult(first=None, last=None, max=None, end=True, items=[])

        response_body = request.get_data()
        response = json.loads(response_body)

        result = response["result"]
        items = result.get("items")
        if items is None:
            yield MuclumbusResult(first=None, last=None, max=None, end=True, items=[])

        results: list[MuclumbusItem] = []
        for item in items:
            try:
                anonymity_mode = AnonymityMode(item["anonymity_mode"])
            except (ValueError, KeyError):
                anonymity_mode = AnonymityMode.UNKNOWN

            results.append(
                MuclumbusItem(
                    jid=item["address"],
                    name=item["name"] or "",
                    nusers=str(item["nusers"] or ""),
                    description=item["description"] or "",
                    language=item["language"] or "",
                    is_open=item["is_open"],
                    anonymity_mode=anonymity_mode,
                )
            )

        yield MuclumbusResult(
            first=None,
            last=result["last"],
            max=None,
            end=not result["more"],
            items=results,
        )


def _make_parameter_request(jid: str) -> Iq:
    query = Iq(to=jid, typ="get")
    query.addChild(node=Node("search", attrs={"xmlns": Namespace.MUCLUMBUS}))
    return query


def _make_search_query(
    jid: str, dataform, items_per_page: int = 50, after: str | None = None
) -> Iq:
    search = Node("search", attrs={"xmlns": Namespace.MUCLUMBUS})
    search.addChild(node=dataform)
    rsm = search.addChild("set", namespace=Namespace.RSM)
    rsm.addChild("max").setData(items_per_page)
    if after is not None:
        rsm.addChild("after").setData(after)
    query = Iq(to=jid, typ="get")
    query.addChild(node=search)
    return query



from test.lib.util import StanzaHandlerTest

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import MoodData
from nbxmpp.structs import PubSubEventData
from nbxmpp.structs import StanzaHandler


class MoodTest(StanzaHandlerTest):

    def test_mood_parsing(self):
        def _on_message(_con, _stanza, properties):

            data = MoodData(mood="annoyed", text="curse my nurse!")

            pubsub_event = PubSubEventData(
                node="http://jabber.org/protocol/mood",
                id="a475804a-0f9c-11dc-98a8-001143d5d5db",
                item=None,
                data=data,
                deleted=False,
                retracted=False,
                purged=False,
            )

            # We cant compare Node objects
            pubsub_event_ = properties.pubsub_event._replace(item=None)
            self.assertEqual(pubsub_event, pubsub_event_)

        event = """
            <message from='test@test.test'>
                <event xmlns='http://jabber.org/protocol/pubsub#event'>
                    <items node='http://jabber.org/protocol/mood'>
                        <item id='a475804a-0f9c-11dc-98a8-001143d5d5db'>
                            <mood xmlns='http://jabber.org/protocol/mood'>
                                <annoyed/>
                                <text>curse my nurse!</text>
                            </mood>
                        </item>
                    </items>
                </event>
            </message>
        """

        self.dispatcher.register_handler(
            StanzaHandler(
                name="message", callback=_on_message, ns=Namespace.PUBSUB_EVENT
            )
        )

        self.dispatcher.process_data(event)



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Literal
from typing import TYPE_CHECKING

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.protocol import Node
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import OMEMOBundle
from nbxmpp.structs import OMEMOMessage
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task
from nbxmpp.util import b64decode
from nbxmpp.util import b64encode
from nbxmpp.util import from_xs_boolean

if TYPE_CHECKING:
    from nbxmpp.client import Client


class OMEMO(BaseModule):

    _depends = {
        "publish": "PubSub",
        "request_items": "PubSub",
    }

    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_omemo_devicelist,
                ns=Namespace.PUBSUB_EVENT,
                priority=16,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_omemo_message,
                ns=Namespace.OMEMO_TEMP,
                priority=7,
            ),
        ]

    def _process_omemo_message(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        try:
            properties.omemo = _parse_omemo_message(stanza)
            self._log.info("Received message")
        except MalformedStanzaError as error:
            self._log.warning(error)
            self._log.warning(stanza)
            return

    def _process_omemo_devicelist(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.OMEMO_TEMP_DL:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        try:
            devices = _parse_devicelist(item)
        except MalformedStanzaError as error:
            self._log.warning(error)
            self._log.warning(stanza)
            raise NodeProcessed

        if not devices:
            self._log.info(
                "Received OMEMO devicelist: %s - no devices set", properties.jid
            )
            return

        pubsub_event = properties.pubsub_event._replace(data=devices)
        self._log.info("Received OMEMO devicelist: %s - %s", properties.jid, devices)

        properties.pubsub_event = pubsub_event

    @iq_request_task
    def set_devicelist(self, devicelist: set[int] | None = None, public: bool = True):
        task = yield

        access_model = "open" if public else "presence"

        options = {
            "pubsub#persist_items": "true",
            "pubsub#access_model": access_model,
        }

        result = yield self.publish(
            Namespace.OMEMO_TEMP_DL,
            _make_devicelist(devicelist),
            id_="current",
            options=options,
            force_node_options=True,
        )

        yield finalize(task, result)

    @iq_request_task
    def request_devicelist(self, jid: str | None = None):
        task = yield

        items = yield self.request_items(Namespace.OMEMO_TEMP_DL, max_items=1, jid=jid)

        raise_if_error(items)

        if not items:
            yield task.set_result(None)

        yield _parse_devicelist(items[0])

    @iq_request_task
    def set_bundle(self, bundle: OMEMOBundle, device_id: int, public: bool = True):
        task = yield

        access_model = "open" if public else "presence"

        options = {
            "pubsub#persist_items": "true",
            "pubsub#access_model": access_model,
        }

        result = yield self.publish(
            f"{Namespace.OMEMO_TEMP_BUNDLE}:{device_id}",
            _make_bundle(bundle),
            id_="current",
            options=options,
            force_node_options=True,
        )

        yield finalize(task, result)

    @iq_request_task
    def request_bundle(self, jid: str, device_id: int):
        task = yield

        items = yield self.request_items(
            f"{Namespace.OMEMO_TEMP_BUNDLE}:{device_id}", max_items=1, jid=jid
        )

        raise_if_error(items)

        if not items:
            yield task.set_result(None)

        yield _parse_bundle(items[0], device_id)


def _parse_omemo_message(stanza: Message) -> OMEMOMessage:
    """
    <message>
      <encrypted xmlns='eu.siacs.conversations.axolotl'>
        <header sid='27183'>
          <key rid='31415'>BASE64ENCODED...</key>
          <key prekey="true" rid='12321'>BASE64ENCODED...</key>
          <!-- ... -->
          <iv>BASE64ENCODED...</iv>
        </header>
        <payload>BASE64ENCODED</payload>
      </encrypted>
      <store xmlns='urn:xmpp:hints'/>
    </message>
    """
    encrypted = stanza.getTag("encrypted", namespace=Namespace.OMEMO_TEMP)
    if encrypted is None:
        raise MalformedStanzaError("No encrypted node found", stanza)

    header = encrypted.getTag("header")
    if header is None:
        raise MalformedStanzaError("header node not found", stanza)

    try:
        sid = int(header.getAttr("sid"))
    except Exception:
        raise MalformedStanzaError("sid attr not found", stanza)

    iv_node = header.getTag("iv")
    try:
        iv = b64decode(iv_node.getData())
    except Exception as error:
        raise MalformedStanzaError("failed to decode iv: %s" % error, stanza)

    payload = None
    payload_node = encrypted.getTag("payload")
    if payload_node is not None:
        try:
            payload = b64decode(payload_node.getData())
        except Exception as error:
            raise MalformedStanzaError("failed to decode payload: %s" % error, stanza)

    key_nodes = header.getTags("key")
    if not key_nodes:
        raise MalformedStanzaError("no keys found", stanza)

    keys: dict[int, tuple[bytes, bool]] = {}
    for kn in key_nodes:
        rid = kn.getAttr("rid")
        if rid is None:
            raise MalformedStanzaError("rid not found", stanza)

        prekey = kn.getAttr("prekey")
        if prekey is None:
            prekey = False
        else:
            try:
                prekey = from_xs_boolean(prekey)
            except ValueError as error:
                raise MalformedStanzaError(error, stanza)

        try:
            keys[int(rid)] = (b64decode(kn.getData()), prekey)
        except Exception as error:
            raise MalformedStanzaError("failed to decode key: %s" % error, stanza)

    return OMEMOMessage(sid=sid, iv=iv, keys=keys, payload=payload)


def _parse_bundle(item: Node | None, device_id: int) -> OMEMOBundle:
    """
    <item id='current'>
      <bundle xmlns='eu.siacs.conversations.axolotl'>
        <signedPreKeyPublic signedPreKeyId='1'>
          BASE64ENCODED...
        </signedPreKeyPublic>
        <signedPreKeySignature>
          BASE64ENCODED...
        </signedPreKeySignature>
        <identityKey>
          BASE64ENCODED...
        </identityKey>
        <prekeys>
          <preKeyPublic preKeyId='1'>
            BASE64ENCODED...
          </preKeyPublic>
          <preKeyPublic preKeyId='2'>
           BASE64ENCODED...
          </preKeyPublic>
          <preKeyPublic preKeyId='3'>
            BASE64ENCODED...
          </preKeyPublic>
          <!-- ... -->
        </prekeys>
      </bundle>
    </item>
    """
    if item is None:
        raise MalformedStanzaError("No item in node found", item)

    bundle = item.getTag("bundle", namespace=Namespace.OMEMO_TEMP)
    if bundle is None:
        raise MalformedStanzaError("No bundle node found", item)

    result = {}
    signed_prekey_node = bundle.getTag("signedPreKeyPublic")
    try:
        result["spk"] = {"key": b64decode(signed_prekey_node.getData())}
    except Exception as error:
        error = "Failed to decode signedPreKeyPublic: %s" % error
        raise MalformedStanzaError(error, item)

    signed_prekey_id = signed_prekey_node.getAttr("signedPreKeyId")
    try:
        result["spk"]["id"] = int(signed_prekey_id)
    except Exception as error:
        raise MalformedStanzaError("Invalid signedPreKeyId: %s" % error, item)

    signed_signature_node = bundle.getTag("signedPreKeySignature")
    try:
        result["spk_signature"] = b64decode(signed_signature_node.getData())
    except Exception as error:
        error = "Failed to decode signedPreKeySignature: %s" % error
        raise MalformedStanzaError(error, item)

    identity_key_node = bundle.getTag("identityKey")
    try:
        result["ik"] = b64decode(identity_key_node.getData())
    except Exception as error:
        error = "Failed to decode IdentityKey: %s" % error
        raise MalformedStanzaError(error, item)

    prekeys = bundle.getTag("prekeys")
    if prekeys is None or not prekeys.getChildren():
        raise MalformedStanzaError("No prekeys node found", item)

    result["otpks"] = []
    for prekey in prekeys.getChildren():
        try:
            id_ = int(prekey.getAttr("preKeyId"))
        except Exception as error:
            raise MalformedStanzaError("Invalid prekey: %s" % error, item)

        try:
            key = b64decode(prekey.getData())
        except Exception as error:
            raise MalformedStanzaError(
                "Failed to decode preKeyPublic: %s" % error, item
            )

        result["otpks"].append({"key": key, "id": id_})

    return OMEMOBundle(**result, device_id=device_id, namespace=Namespace.OMEMO_TEMP)


def _make_bundle(bundle: OMEMOBundle) -> Node:
    """
    <publish node='eu.siacs.conversations.axolotl.bundles:31415'>
      <item id='current'>
        <bundle xmlns='eu.siacs.conversations.axolotl'>
          <signedPreKeyPublic signedPreKeyId='1'>
            BASE64ENCODED...
          </signedPreKeyPublic>
          <signedPreKeySignature>
            BASE64ENCODED...
          </signedPreKeySignature>
          <identityKey>
            BASE64ENCODED...
          </identityKey>
          <prekeys>
            <preKeyPublic preKeyId='1'>
              BASE64ENCODED...
            </preKeyPublic>
            <preKeyPublic preKeyId='2'>
              BASE64ENCODED...
            </preKeyPublic>
            <preKeyPublic preKeyId='3'>
              BASE64ENCODED...
            </preKeyPublic>
            <!-- ... -->
          </prekeys>
        </bundle>
      </item>
    </publish>
    """
    bundle_node = Node("bundle", attrs={"xmlns": Namespace.OMEMO_TEMP})
    prekey_pub_node = bundle_node.addChild(
        "signedPreKeyPublic", attrs={"signedPreKeyId": bundle.spk["id"]}
    )
    prekey_pub_node.addData(b64encode(bundle.spk["key"]))

    prekey_sig_node = bundle_node.addChild("signedPreKeySignature")
    prekey_sig_node.addData(b64encode(bundle.spk_signature))

    identity_key_node = bundle_node.addChild("identityKey")
    identity_key_node.addData(b64encode(bundle.ik))

    prekeys = bundle_node.addChild("prekeys")
    for key in bundle.otpks:
        pre_key_public = prekeys.addChild("preKeyPublic", attrs={"preKeyId": key["id"]})
        pre_key_public.addData(b64encode(key["key"]))
    return bundle_node


def _make_devicelist(devicelist: set[int] | None) -> Node:
    if devicelist is None:
        devicelist = set()

    devicelist_node = Node("list", attrs={"xmlns": Namespace.OMEMO_TEMP})
    for device in devicelist:
        devicelist_node.addChild("device").setAttr("id", device)

    return devicelist_node


def _parse_devicelist(item: Node) -> list[int]:
    """
    <items node='eu.siacs.conversations.axolotl.devicelist'>
      <item id='current'>
        <list xmlns='eu.siacs.conversations.axolotl'>
          <device id='12345' />
          <device id='4223' />
        </list>
      </item>
    </items>
    """
    list_node = item.getTag("list", namespace=Namespace.OMEMO_TEMP)
    if list_node is None:
        raise MalformedStanzaError("No list node found", item)

    if not list_node.getChildren():
        return []

    result: list[int] = []
    devices_nodes = list_node.getChildren()
    for dn in devices_nodes:
        _id = dn.getAttr("id")
        if _id:
            result.append(int(_id))

    return result


def create_omemo_message(
    stanza: Message,
    omemo_message: OMEMOMessage,
    store_hint: bool = True,
    node_whitelist: list[tuple[str, str]] | None = None,
) -> None:
    """
    <message>
      <encrypted xmlns='eu.siacs.conversations.axolotl'>
        <header sid='27183'>
          <key rid='31415'>BASE64ENCODED...</key>
          <key prekey="true" rid='12321'>BASE64ENCODED...</key>
          <!-- ... -->
          <iv>BASE64ENCODED...</iv>
        </header>
        <payload>BASE64ENCODED</payload>
      </encrypted>
      <store xmlns='urn:xmpp:hints'/>
    </message>
    """

    if node_whitelist is not None:
        cleanup_stanza(stanza, node_whitelist)

    encrypted = Node("encrypted", attrs={"xmlns": Namespace.OMEMO_TEMP})
    header = Node("header", attrs={"sid": omemo_message.sid})
    for rid, (key, prekey) in omemo_message.keys.items():
        attrs = {"rid": rid}
        if prekey:
            attrs["prekey"] = "true"
        child = header.addChild("key", attrs=attrs)
        child.addData(b64encode(key))

    header.addChild("iv").addData(b64encode(omemo_message.iv))
    encrypted.addChild(node=header)

    payload = encrypted.addChild("payload")
    payload.addData(b64encode(omemo_message.payload))

    stanza.addChild(node=encrypted)

    stanza.addChild(
        node=Node(
            "encryption",
            attrs={
                "xmlns": Namespace.EME,
                "name": "OMEMO",
                "namespace": Namespace.OMEMO_TEMP,
            },
        )
    )

    stanza.setBody(
        "You received a message encrypted with "
        "OMEMO but your client doesn't support OMEMO."
    )

    if store_hint:
        stanza.addChild(node=Node("store", attrs={"xmlns": Namespace.HINTS}))


def get_key_transport_message(
    typ: Literal["chat", "groupchat"], jid: str, omemo_message: OMEMOMessage
) -> Message:
    message = Message(typ=typ, to=jid)

    encrypted = Node("encrypted", attrs={"xmlns": Namespace.OMEMO_TEMP})
    header = Node("header", attrs={"sid": omemo_message.sid})
    for rid, (key, prekey) in omemo_message.keys.items():
        attrs = {"rid": rid}
        if prekey:
            attrs["prekey"] = "true"
        child = header.addChild("key", attrs=attrs)
        child.addData(b64encode(key))

    header.addChild("iv").addData(b64encode(omemo_message.iv))
    encrypted.addChild(node=header)

    message.addChild(node=encrypted)
    return message


def cleanup_stanza(stanza: Message, node_whitelist: list[tuple[str, str]]) -> None:
    whitelisted_nodes: list[Node] = []
    for tag, ns in node_whitelist:
        node = stanza.getTag(tag, namespace=ns)
        if node is not None:
            whitelisted_nodes.append(node)

    for node in list(stanza.getChildren()):
        stanza.delChild(node)

    for node in whitelisted_nodes:
        stanza.addChild(node=node)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Presence
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Idle(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="presence",
                callback=self._process_idle,
                ns=Namespace.IDLE,
                priority=15,
            )
        ]

    def _process_idle(
        self, _client: Client, stanza: Presence, properties: PresenceProperties
    ) -> None:
        idle_tag = stanza.getTag("idle", namespace=Namespace.IDLE)
        if idle_tag is None:
            return

        since = idle_tag.getAttr("since")
        if since is None:
            self._log.warning("No since attr in idle node")
            self._log.warning(stanza)
            return

        timestamp = parse_datetime(since, convert="utc", epoch=True)
        if timestamp is None:
            self._log.warning("Invalid timestamp received: %s", since)
            self._log.warning(stanza)

        properties.idle_timestamp = timestamp



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import Optional
from typing import TYPE_CHECKING

import binascii
import hashlib
import hmac
import logging
import os
from hashlib import pbkdf2_hmac

from gi.repository import Gio

from nbxmpp.const import StreamState
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Node
from nbxmpp.protocol import Protocol
from nbxmpp.protocol import SASL_AUTH_MECHS
from nbxmpp.protocol import SASL_ERROR_CONDITIONS
from nbxmpp.structs import ChannelBindingData
from nbxmpp.util import b64decode
from nbxmpp.util import b64encode
from nbxmpp.util import LogAdapter

if TYPE_CHECKING:
    from nbxmpp.client import Client
    from nbxmpp.protocol import Features

log = logging.getLogger("nbxmpp.sasl")

try:
    gssapi = __import__("gssapi")
    gssapi_available = True
except (ImportError, OSError) as error:
    log.info("GSSAPI not available: %s", error)
    gssapi_available = False

GSSAPI_AVAILABLE = gssapi_available


class SASL:
    """
    Implements SASL authentication.
    """

    def __init__(self, client: Client) -> None:
        self._client = client

        self._password: str | None = None

        self._mechanism_classes = {
            "ANONYMOUS": ANONYMOUS,
            "PLAIN": PLAIN,
            "EXTERNAL": EXTERNAL,
            "GSSAPI": GSSAPI,
            "SCRAM-SHA-1": SCRAM_SHA_1,
            "SCRAM-SHA-1-PLUS": SCRAM_SHA_1_PLUS,
            "SCRAM-SHA-256": SCRAM_SHA_256,
            "SCRAM-SHA-256-PLUS": SCRAM_SHA_256_PLUS,
            "SCRAM-SHA-512": SCRAM_SHA_512,
            "SCRAM-SHA-512-PLUS": SCRAM_SHA_512_PLUS,
        }

        self._allowed_mechs: set[str] | None = None
        self._enabled_mechs: set[str] | None = None
        self._sasl_ns: str | None = None
        self._mechanism: BaseMechanism | None = None
        self._error: tuple[str | None, str | None] | None = None

        self._log = LogAdapter(log, {"context": client.log_context})

    @property
    def error(self) -> tuple[str | None, str | None] | None:
        return self._error

    def is_sasl2(self) -> bool:
        assert self._sasl_ns is not None
        return self._sasl_ns == Namespace.SASL2

    def set_password(self, password: str | None):
        self._password = password

    @property
    def password(self) -> str | None:
        return self._password

    def delegate(self, stanza: Protocol) -> None:
        if stanza.getNamespace() != self._sasl_ns:
            return

        if stanza.getName() == "challenge":
            self._on_challenge(stanza)
        elif stanza.getName() == "failure":
            self._on_failure(stanza)
        elif stanza.getName() == "success":
            self._on_success(stanza)

    def _get_channel_binding_data(
        self, features: Features
    ) -> Optional[ChannelBindingData]:
        if self._client.tls_version != Gio.TlsProtocolVersion.TLS_1_3:
            return None

        binding_type = features.get_channel_binding_type()
        if binding_type is None:
            return None

        channel_binding_data = self._client.get_channel_binding_data(binding_type)
        if channel_binding_data is None:
            return None

        return ChannelBindingData(binding_type, channel_binding_data)

    def start_auth(self, features: Features) -> None:
        self._mechanism = None
        self._allowed_mechs = self._client.mechs
        self._enabled_mechs = self._allowed_mechs

        self._sasl_ns = Namespace.SASL
        if features.has_sasl_2():
            self._sasl_ns = Namespace.SASL2

        self._log.info("Using %s", self._sasl_ns)

        self._error = None

        channel_binding_data = None
        # Segfaults see https://gitlab.gnome.org/GNOME/pygobject/-/issues/603
        # So for now channel binding is deactivated
        # channel_binding_data = self._get_channel_binding_data(features)
        if channel_binding_data is None:
            self._enabled_mechs.discard("SCRAM-SHA-1-PLUS")
            self._enabled_mechs.discard("SCRAM-SHA-256-PLUS")
            self._enabled_mechs.discard("SCRAM-SHA-512-PLUS")

        if not GSSAPI_AVAILABLE:
            self._enabled_mechs.discard("GSSAPI")

        available_mechs = features.get_mechs() & self._enabled_mechs
        self._log.info("Available mechanisms: %s", available_mechs)

        domain_based_name = features.get_domain_based_name()
        if domain_based_name is not None:
            self._log.info("Found domain based name: %s", domain_based_name)

        if not available_mechs:
            self._log.error("No available auth mechanisms found")
            self._abort_auth("invalid-mechanism")
            return

        chosen_mechanism = None
        for mech in SASL_AUTH_MECHS:
            if mech in available_mechs:
                chosen_mechanism = mech
                break

        if chosen_mechanism is None:
            self._log.error("No available auth mechanisms found")
            self._abort_auth("invalid-mechanism")
            return

        self._log.info("Chosen auth mechanism: %s", chosen_mechanism)

        if chosen_mechanism.startswith(("SCRAM", "PLAIN")):
            if not self._password:
                self._on_sasl_finished(False, "no-password")
                return

        mech_class = self._mechanism_classes[chosen_mechanism]
        self._mechanism = mech_class(
            self._client.username,
            self._password,
            domain_based_name or self._client.domain,
        )

        if isinstance(self._mechanism, SCRAM) and channel_binding_data is not None:
            self._mechanism.set_channel_binding_data(channel_binding_data)

        try:
            self._send_initiate()
        except AuthFail as error:
            self._log.error(error)
            self._abort_auth()
            return

    def _send_initiate(self) -> None:
        assert self._mechanism is not None
        data = self._mechanism.get_initiate_data()
        nonza = get_initiate_nonza(self._sasl_ns, self._mechanism.name, data)
        self._client.send_nonza(nonza)

    def _on_challenge(self, stanza: Protocol) -> None:
        assert self._mechanism is not None
        try:
            data = self._mechanism.get_response_data(stanza.getData())
        except AttributeError:
            self._log.info("Mechanism has no response method")
            self._abort_auth()
            return

        except AuthFail as error:
            self._log.error(error)
            self._abort_auth()
            return

        nonza = get_response_nonza(self._sasl_ns, data)
        self._client.send_nonza(nonza)

    def _on_success(self, stanza: Protocol) -> None:
        self._log.info("Successfully authenticated with remote server")
        data = get_success_data(stanza, self._sasl_ns)
        try:
            self._mechanism.validate_success_data(data)
        except Exception as error:
            self._log.error("Unable to validate success data: %s", error)
            self._abort_auth()
            return

        self._log.info("Validated success data")

        self._on_sasl_finished(True, None, None)

    def _on_failure(self, stanza: Protocol) -> None:
        text = stanza.getTagData("text")
        reason = "not-authorized"
        childs = stanza.getChildren()
        for child in childs:
            name = child.getName()
            if name == "text":
                continue
            if name in SASL_ERROR_CONDITIONS:
                reason = name
                break

        self._log.info("Failed SASL authentification: %s %s", reason, text)
        self._abort_auth(reason, text)

    def _abort_auth(
        self, reason: str = "malformed-request", text: str | None = None
    ) -> None:
        node = Node("abort", attrs={"xmlns": self._sasl_ns})
        self._client.send_nonza(node)
        self._on_sasl_finished(False, reason, text)

    def _on_sasl_finished(
        self, successful: bool, reason: str | None, text: str | None = None
    ) -> None:
        if not successful:
            self._error = (reason, text)
            self._client.set_state(StreamState.AUTH_FAILED)
        else:
            self._client.set_state(StreamState.AUTH_SUCCESSFUL)


def get_initiate_nonza(ns: str, mechanism: str, data: str | None) -> Node:

    if ns == Namespace.SASL:
        node = Node("auth", attrs={"xmlns": ns, "mechanism": mechanism})
        if data is not None:
            node.setData(data)

    else:
        node = Node("authenticate", attrs={"xmlns": ns, "mechanism": mechanism})
        if data is not None:
            node.setTagData("initial-response", data)

    return node


def get_response_nonza(ns: str, data: str) -> Node:
    return Node("response", attrs={"xmlns": ns}, payload=[data])


def get_success_data(stanza: Protocol, ns: str) -> str | None:
    if ns == Namespace.SASL2:
        return stanza.getTagData("additional-data")
    return stanza.getData()


class BaseMechanism:

    name: str

    def __init__(self, username: str, password: str, domain: str) -> None:
        self._username = username
        self._password = password
        self._domain = domain

    def get_initiate_data(self) -> str | None:
        raise NotImplementedError

    def get_response_data(self, data: str) -> str:
        raise NotImplementedError

    def validate_success_data(self, _data: str) -> None:
        return None


class PLAIN(BaseMechanism):

    name = "PLAIN"

    def get_initiate_data(self) -> str:
        return b64encode("\x00%s\x00%s" % (self._username, self._password))


class EXTERNAL(BaseMechanism):

    name = "EXTERNAL"

    def get_initiate_data(self) -> str:
        return b64encode("%s@%s" % (self._username, self._domain))


class ANONYMOUS(BaseMechanism):

    name = "ANONYMOUS"

    def get_initiate_data(self) -> None:
        return None


class GSSAPI(BaseMechanism):

    # See https://tools.ietf.org/html/rfc4752#section-3.1

    name = "GSSAPI"

    def get_initiate_data(self) -> str:
        service = gssapi.Name(
            "xmpp@%s" % self._domain, name_type=gssapi.NameType.hostbased_service
        )
        try:
            self.ctx = gssapi.SecurityContext(
                name=service, usage="initiate", flags=gssapi.RequirementFlag.integrity
            )
            token = self.ctx.step()
        except (gssapi.exceptions.GeneralError, gssapi.raw.misc.GSSError) as e:
            raise AuthFail(e)

        return b64encode(token)

    def get_response_data(self, data: str | bytes) -> str:
        byte_data = b64decode(data)
        try:
            if not self.ctx.complete:
                output_token = self.ctx.step(byte_data)
            else:
                _result = self.ctx.unwrap(byte_data)
                # TODO(jelmer): Log result.message
                data = b"\x00\x00\x00\x00" + bytes(self.ctx.initiator_name)
                output_token = self.ctx.wrap(data, False).message
        except (gssapi.exceptions.GeneralError, gssapi.raw.misc.GSSError) as e:
            raise AuthFail(e)

        return b64encode(output_token)


class SCRAM(BaseMechanism):

    name = ""
    _hash_method = ""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        BaseMechanism.__init__(self, *args, **kwargs)
        self._channel_binding_data: ChannelBindingData | None = None
        self._gs2_header = "n,,"
        self._client_nonce = "%x" % int(binascii.hexlify(os.urandom(24)), 16)
        self._client_first_message_bare = None
        self._server_signature = None

    def set_channel_binding_data(self, data: ChannelBindingData) -> None:
        self._channel_binding_data = data
        self._gs2_header = f"p={data.type},,"

    @property
    def nonce_length(self) -> int:
        return len(self._client_nonce)

    @property
    def _b64_channel_binding_data(self) -> str:
        if self.name.endswith("PLUS"):
            assert self._channel_binding_data is not None
            return b64encode(
                b"%s%s" % (self._gs2_header.encode(), self._channel_binding_data.data)
            )
        return b64encode(self._gs2_header)

    @staticmethod
    def _scram_parse(scram_data: str) -> dict[str, str]:
        return dict(s.split("=", 1) for s in scram_data.split(","))

    def get_initiate_data(self) -> str:
        self._client_first_message_bare = "n=%s,r=%s" % (
            self._username,
            self._client_nonce,
        )

        client_first_message = "%s%s" % (
            self._gs2_header,
            self._client_first_message_bare,
        )

        return b64encode(client_first_message)

    def get_response_data(self, data: str | bytes) -> str:
        server_first_message = b64decode(data).decode()
        challenge = self._scram_parse(server_first_message)

        client_nonce = challenge["r"][: self.nonce_length]
        if client_nonce != self._client_nonce:
            raise AuthFail("Invalid client nonce received from server")

        salt = b64decode(challenge["s"])
        iteration_count = int(challenge["i"])

        if iteration_count < 4096:
            raise AuthFail("Salt iteration count to low: %s" % iteration_count)

        salted_password = pbkdf2_hmac(
            self._hash_method, self._password.encode("utf8"), salt, iteration_count
        )

        client_final_message_wo_proof = "c=%s,r=%s" % (
            self._b64_channel_binding_data,
            challenge["r"],
        )

        client_key = self._hmac(salted_password, "Client Key")
        stored_key = self._h(client_key)
        auth_message = "%s,%s,%s" % (
            self._client_first_message_bare,
            server_first_message,
            client_final_message_wo_proof,
        )
        client_signature = self._hmac(stored_key, auth_message)
        client_proof = self._xor(client_key, client_signature)

        client_finale_message = "c=%s,r=%s,p=%s" % (
            self._b64_channel_binding_data,
            challenge["r"],
            b64encode(client_proof),
        )

        server_key = self._hmac(salted_password, "Server Key")
        self._server_signature = self._hmac(server_key, auth_message)

        return b64encode(client_finale_message)

    def validate_success_data(self, data: str) -> None:
        server_last_message = b64decode(data).decode()
        success = self._scram_parse(server_last_message)
        server_signature = b64decode(success["v"])
        if server_signature != self._server_signature:
            raise AuthFail("Invalid server signature")

    def _hmac(self, key: bytes, message: str) -> bytes:
        return hmac.new(
            key=key, msg=message.encode(), digestmod=self._hash_method
        ).digest()

    @staticmethod
    def _xor(x: bytes, y: bytes) -> bytes:
        return bytes([px ^ py for px, py in zip(x, y, strict=False)])

    def _h(self, data: bytes) -> bytes:
        return hashlib.new(self._hash_method, data).digest()


class SCRAM_SHA_1(SCRAM):

    name = "SCRAM-SHA-1"
    _hash_method = "sha1"


class SCRAM_SHA_1_PLUS(SCRAM_SHA_1):

    name = "SCRAM-SHA-1-PLUS"


class SCRAM_SHA_256(SCRAM):

    name = "SCRAM-SHA-256"
    _hash_method = "sha256"


class SCRAM_SHA_256_PLUS(SCRAM_SHA_256):

    name = "SCRAM-SHA-256-PLUS"


class SCRAM_SHA_512(SCRAM):

    name = "SCRAM-SHA-512"
    _hash_method = "sha512"


class SCRAM_SHA_512_PLUS(SCRAM_SHA_512):

    name = "SCRAM-SHA-512-PLUS"


class AuthFail(Exception):
    pass



# Copyright (C) 2006-2007 Tomasz Melcer <liori AT exroot.org>
# Copyright (C) 2006-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2007 Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

# XEP-0004: Data Forms

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import Union

from collections.abc import Iterator

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.simplexml import Node

FieldT = Union[
    "BooleanField",
    "StringField",
    "JidMultiField",
    "JidSingleField",
    "ListMultiField",
    "ListSingleField",
    "TextMultiField",
]


# exceptions used in this module
class Error(Exception):
    pass


# when we get nbxmpp.Node which we do not understand
class UnknownDataForm(Error):
    pass


# when we get nbxmpp.Node which contains bad fields
class WrongFieldValue(Error):
    pass


# helper class to change class of already existing object
class ExtendedNode(Node):
    @classmethod
    def __new__(cls, *args: Any, **kwargs: Any):
        if "extend" not in kwargs or not kwargs["extend"]:
            return object.__new__(cls)

        extend = kwargs["extend"]
        assert issubclass(cls, extend.__class__)
        extend.__class__ = cls
        return extend


# helper to create fields from scratch
def create_field(typ: str, **attrs: Any) -> FieldT:
    """Helper function to create a field of given type."""
    field: FieldT = {
        "boolean": BooleanField,
        "fixed": StringField,
        "hidden": StringField,
        "text-private": StringField,
        "text-single": StringField,
        "jid-multi": JidMultiField,
        "jid-single": JidSingleField,
        "list-multi": ListMultiField,
        "list-single": ListSingleField,
        "text-multi": TextMultiField,
    }[typ](typ=typ, **attrs)
    return field


def extend_field(node: Node) -> FieldT:
    """
    Helper function to extend a node to field of appropriate type
    """
    # when validation (XEP-122) will go in, we could have another classes
    # like DateTimeField - so that dicts in create_field() and
    # extend_field() will be different...
    typ = node.getAttr("type")
    field = {
        "boolean": BooleanField,
        "fixed": StringField,
        "hidden": StringField,
        "text-private": StringField,
        "text-single": StringField,
        "jid-multi": JidMultiField,
        "jid-single": JidSingleField,
        "list-multi": ListMultiField,
        "list-single": ListSingleField,
        "text-multi": TextMultiField,
    }
    if typ not in field:
        typ = "text-single"
    return field[typ](extend=node)


def extend_form(node: Node) -> SimpleDataForm | MultipleDataForm:
    """
    Helper function to extend a node to form of appropriate type
    """
    if node.getTag("reported") is not None:
        return MultipleDataForm(extend=node)
    return SimpleDataForm(extend=node)


class DataField(ExtendedNode):
    """
    Keeps data about one field - var, field type, labels, instructions... Base
    class for different kinds of fields. Use create_field() function to
    construct one of these
    """

    def __init__(
        self,
        typ: str | None = None,
        var: str | None = None,
        value: str | None = None,
        label: str | None = None,
        desc: str | None = None,
        required: bool = False,
        options: list[tuple[str, str]] | None = None,
        extend: Node | None = None,
    ) -> None:

        if extend is None:
            ExtendedNode.__init__(self, "field")

            self.type_ = typ
            self.var = var
            if value is not None:
                self.value = value
            if label is not None:
                self.label = label
            if desc is not None:
                self.desc = desc
            self.required = required
            self.options = options

    @property
    def type_(self) -> str:
        """
        Type of field. Recognized values are: 'boolean', 'fixed', 'hidden',
        'jid-multi', 'jid-single', 'list-multi', 'list-single', 'text-multi',
        'text-private', 'text-single'. If you set this to something different,
        DataField will store given name, but treat all data as text-single
        """
        type_ = self.getAttr("type")
        if type_ is None:
            return "text-single"
        return type_

    @type_.setter
    def type_(self, value: str) -> None:
        assert isinstance(value, str)
        self.setAttr("type", value)

    @property
    def var(self) -> str:
        """
        Field identifier
        """
        return self.getAttr("var")

    @var.setter
    def var(self, value: str) -> None:
        assert isinstance(value, str)
        self.setAttr("var", value)

    @var.deleter
    def var(self) -> None:
        self.delAttr("var")

    @property
    def label(self) -> str:
        """
        Human-readable field name
        """
        label_ = self.getAttr("label")
        if not label_:
            label_ = self.var
        return label_

    @label.setter
    def label(self, value: str) -> None:
        assert isinstance(value, str)
        self.setAttr("label", value)

    @label.deleter
    def label(self) -> None:
        if self.getAttr("label"):
            self.delAttr("label")

    @property
    def description(self) -> str:
        """
        Human-readable description of field meaning
        """
        return self.getTagData("desc") or ""

    @description.setter
    def description(self, value: str) -> None:
        assert isinstance(value, str)
        if value == "":
            del self.description
        else:
            self.setTagData("desc", value)

    @description.deleter
    def description(self) -> None:
        desc = self.getTag("desc")
        if desc is not None:
            self.delChild(desc)

    @property
    def required(self) -> bool:
        """
        Controls whether this field required to fill. Boolean
        """
        return bool(self.getTag("required"))

    @required.setter
    def required(self, value: bool) -> None:
        required = self.getTag("required")
        if required and not value:
            self.delChild(required)
        elif not required and value:
            self.addChild("required")

    @property
    def media(self) -> Media | None:
        """
        Media data
        """
        media = self.getTag("media", namespace=Namespace.DATA_MEDIA)
        if media:
            return Media(media)
        return None

    @media.setter
    def media(self, value: Media) -> None:
        del self.media
        self.addChild(node=value)

    @media.deleter
    def media(self) -> None:
        media = self.getTag("media")
        if media is not None:
            self.delChild(media)

    def is_valid(self) -> tuple[bool, str]:
        return True, ""


class Uri(Node):
    def __init__(self, uri_tag: Node) -> None:
        Node.__init__(self, node=uri_tag)

    @property
    def type_(self) -> str | None:
        """
        uri type
        """
        return self.getAttr("type")

    @type_.setter
    def type_(self, value: str) -> None:
        self.setAttr("type", value)

    @type_.deleter
    def type_(self) -> None:
        self.delAttr("type")

    @property
    def uri_data(self) -> str:
        """
        uri data
        """
        return self.getData()

    @uri_data.setter
    def uri_data(self, value: str) -> None:
        self.setData(value)

    @uri_data.deleter
    def uri_data(self) -> None:
        self.setData(None)


class Media(Node):
    def __init__(self, media_tag: Node) -> None:
        Node.__init__(self, node=media_tag)

    @property
    def uris(self) -> list[Uri]:
        """
        URIs of the media element.
        """
        return list(map(Uri, self.getTags("uri")))

    @uris.setter
    def uris(self, value: list[Uri]) -> None:
        del self.uris
        for uri in value:
            self.addChild(node=uri)

    @uris.deleter
    def uris(self) -> None:
        for element in self.getTags("uri"):
            self.delChild(element)


class BooleanField(DataField):
    @property
    def value(self) -> bool:
        """
        Value of field. May contain True, False or None
        """
        value = self.getTagData("value")
        if value in ("0", "false"):
            return False
        if value in ("1", "true"):
            return True
        if value is None:
            return False  # default value is False
        raise WrongFieldValue

    @value.setter
    def value(self, value: bool) -> None:
        self.setTagData("value", (value and "1") or "0")

    @value.deleter
    def value(self) -> None:
        value = self.getTag("value")
        if value is not None:
            self.delChild(value)


class StringField(DataField):
    """
    Covers fields of types: fixed, hidden, text-private, text-single
    """

    @property
    def value(self) -> str:
        """
        Value of field. May be any string
        """
        return self.getTagData("value") or ""

    @value.setter
    def value(self, value: str) -> None:
        if value is None:  # type: ignore
            value = ""
        self.setTagData("value", value)

    @value.deleter
    def value(self) -> None:
        try:
            self.delChild(self.getTag("value"))  # type: ignore
        except ValueError:  # if there already were no value tag
            pass

    def is_valid(self) -> tuple[bool, str]:
        if not self.required:
            return True, ""
        if not self.value:
            return False, ""
        return True, ""


class ListField(DataField):
    """
    Covers fields of types: jid-multi, jid-single, list-multi, list-single
    """

    @property
    def options(self) -> list[tuple[str, Any]]:
        """
        Options
        """
        options: list[tuple[str, Any]] = []
        for element in self.getTags("option"):
            value = element.getTagData("value")
            if value is None:
                raise WrongFieldValue
            label = element.getAttr("label")
            if not label:
                label = value
            options.append((label, value))
        return options

    @options.setter
    def options(self, values: list[tuple[Any, str]]) -> None:
        del self.options
        for value, label in values:
            self.addChild("option", {"label": label}).setTagData("value", value)

    @options.deleter
    def options(self) -> None:
        for element in self.getTags("option"):
            self.delChild(element)

    def iter_options(self) -> Iterator[tuple[Any, str]]:
        for element in self.iterTags("option"):
            value = element.getTagData("value")
            if value is None:
                raise WrongFieldValue
            label = element.getAttr("label")
            if not label:
                label = value
            yield (value, label)


class ListSingleField(ListField, StringField):
    """
    Covers list-single field
    """

    def is_valid(self) -> tuple[bool, str]:
        if not self.required:
            return True, ""
        if not self.value:
            return False, ""
        return True, ""


class JidSingleField(ListSingleField):
    """
    Covers jid-single fields
    """

    def is_valid(self) -> tuple[bool, str]:
        if self.value:
            try:
                JID.from_string(self.value)
                return True, ""
            except Exception as error:
                return False, str(error)
        if self.required:
            return False, ""
        return True, ""


class ListMultiField(ListField):
    """
    Covers list-multi fields
    """

    @property
    def values(self) -> list[str]:
        """
        Values held in field
        """
        values: list[str] = []
        for element in self.getTags("value"):
            values.append(element.getData())
        return values

    @values.setter
    def values(self, values: list[Any]) -> None:
        del self.values
        for value in values:
            self.addChild("value").setData(value)

    @values.deleter
    def values(self) -> None:
        for element in self.getTags("value"):
            self.delChild(element)

    def iter_values(self) -> Iterator[str]:
        for element in self.getTags("value"):
            yield element.getData()

    def is_valid(self) -> tuple[bool, str]:
        if not self.required:
            return True, ""
        if not self.values:
            return False, ""
        return True, ""


class JidMultiField(ListMultiField):
    """
    Covers jid-multi fields
    """

    def is_valid(self) -> tuple[bool, str]:
        if self.values:
            for value in self.values:
                try:
                    JID.from_string(value)
                except Exception as error:
                    return False, str(error)
            return True, ""
        if self.required:
            return False, ""
        return True, ""


class TextMultiField(DataField):
    @property
    def value(self) -> str:
        """
        Value held in field
        """
        value = ""
        for element in self.iterTags("value"):
            value += "\n" + element.getData()
        return value[1:]

    @value.setter
    def value(self, value: str) -> None:
        del self.value
        if value == "":
            return
        for line in value.split("\n"):
            self.addChild("value").setData(line)

    @value.deleter
    def value(self) -> None:
        for element in self.getTags("value"):
            self.delChild(element)

    def is_valid(self) -> tuple[bool, str]:
        if not self.required:
            return True, ""
        if not self.value:
            return False, ""
        return True, ""


class DataRecord(ExtendedNode):
    """
    The container for data fields - an xml element which has DataField elements
    as children
    """

    def __init__(
        self,
        fields: list[FieldT] | None = None,
        associated: SimpleDataForm | None = None,
        extend: Node | None = None,
    ) -> None:
        self.associated = associated
        self.vars: dict[str, FieldT] = {}
        if extend is None:
            # we have to build this object from scratch
            Node.__init__(self)

            if fields is not None:
                self.fields = fields
        else:
            # we already have nbxmpp.Node inside - try to convert all
            # fields into DataField objects
            if fields is None:
                for field in self.iterTags("field"):
                    if not isinstance(field, DataField):
                        extend_field(field)
                    self.vars[field.var] = field
            else:
                self.fields = fields

    @property
    def fields(self) -> list[FieldT]:
        """
        List of fields in this record
        """
        return self.getTags("field")

    @fields.setter
    def fields(self, fields: list[FieldT]) -> None:
        del self.fields
        for field in fields:
            if not isinstance(field, DataField):
                extend_field(field)
            self.addChild(node=field)
            self.vars[field.var] = field

    @fields.deleter
    def fields(self) -> None:
        for element in self.getTags("field"):
            self.delChild(element)
            self.vars.clear()

    def iter_fields(self) -> Iterator[FieldT]:
        """
        Iterate over fields in this record. Do not take associated into account
        """
        yield from self.iterTags("field")

    def iter_with_associated(self) -> Iterator[tuple[Node, FieldT]]:
        """
        Iterate over associated, yielding both our field and associated one
        together
        """
        for field in self.associated.iter_fields():
            yield self[field.var], field

    def __getitem__(self, item: str) -> FieldT:
        return self.vars[item]

    def is_valid(self) -> bool:
        return all(field.is_valid()[0] for field in self.iter_fields())

    def is_fake_form(self) -> bool:
        return bool(self.vars.get("fakeform", False))


class DataForm(ExtendedNode):
    def __init__(
        self,
        type_: str | None = None,
        title: str | None = None,
        instructions: str | None = None,
        extend: Node | None = None,
    ) -> None:
        if extend is None:
            # we have to build form from scratch
            Node.__init__(self, "x", attrs={"xmlns": Namespace.DATA})

        if type_ is not None:
            self.type_ = type_
        if title is not None:
            self.title = title
        if instructions is not None:
            self.instructions = instructions

    @property
    def type_(self) -> str | None:
        """
        Type of the form. Must be one of: 'form', 'submit', 'cancel', 'result'.
        'form' - this form is to be filled in; you will be able soon to do:
        filledform = DataForm(replyto=thisform)
        """
        return self.getAttr("type")

    @type_.setter
    def type_(self, type_: Literal["form", "submit", "cancel", "result"]) -> None:
        assert type_ in ("form", "submit", "cancel", "result")
        self.setAttr("type", type_)

    @property
    def title(self) -> str | None:
        """
        Title of the form

        Human-readable, should not contain any \\r\\n.
        """
        return self.getTagData("title")

    @title.setter
    def title(self, title: str) -> None:
        self.setTagData("title", title)

    @title.deleter
    def title(self) -> None:
        try:
            self.delChild("title")
        except ValueError:
            pass

    @property
    def instructions(self) -> str:
        """
        Instructions for this form

        Human-readable, may contain \\r\\n.
        """
        # TODO: the same code is in TextMultiField. join them
        value = ""
        for valuenode in self.getTags("instructions"):
            value += "\n" + valuenode.getData()
        return value[1:]

    @instructions.setter
    def instructions(self, value: str) -> None:
        del self.instructions
        if value == "":
            return
        for line in value.split("\n"):
            self.addChild("instructions").setData(line)

    @instructions.deleter
    def instructions(self) -> None:
        for value in self.getTags("instructions"):
            self.delChild(value)

    @property
    def is_reported(self) -> bool:
        return self.getTag("reported") is not None


class SimpleDataForm(DataForm, DataRecord):
    def __init__(
        self,
        type_: str | None = None,
        title: str | None = None,
        instructions: str | None = None,
        fields: list[FieldT] | None = None,
        extend: SimpleDataForm | Node | None = None,
    ) -> None:
        DataForm.__init__(
            self, type_=type_, title=title, instructions=instructions, extend=extend
        )
        DataRecord.__init__(self, fields=fields, extend=self, associated=self)

    def get_purged(self) -> SimpleDataForm:
        simple_form = SimpleDataForm(extend=self)
        del simple_form.title
        simple_form.instructions = ""
        to_be_removed: list[FieldT] = []
        for field in simple_form.iter_fields():
            if field.required:
                # add <value> if there is not
                if hasattr(field, "value") and not field.value:
                    field.value = ""
                # Keep all required fields
                continue
            if (hasattr(field, "value") and not field.value and field.value != 0) or (
                hasattr(field, "values") and not field.values
            ):
                to_be_removed.append(field)
            else:
                del field.label
                del field.description
                del field.media
        for field in to_be_removed:
            simple_form.delChild(field)
        return simple_form


class MultipleDataForm(DataForm):
    def __init__(
        self,
        type_: str | None = None,
        title: str | None = None,
        instructions: str | None = None,
        items: list[Node] | None = None,
        extend: Node | None = None,
    ) -> None:
        DataForm.__init__(
            self, type_=type_, title=title, instructions=instructions, extend=extend
        )
        # all records, recorded into DataRecords
        if extend is None:
            if items is not None:
                self.items = items
        else:
            # we already have nbxmpp.Node inside - try to convert all
            # fields into DataField objects
            if items is None:
                self.items = list(self.iterTags("item"))
            else:
                for item in self.getTags("item"):
                    self.delChild(item)
                self.items = items
        reported_tag = self.getTag("reported")
        self.reported = DataRecord(extend=reported_tag)

    @property
    def items(self) -> list[Node]:
        """
        A list of all records
        """
        return list(self.iter_records())

    @items.setter
    def items(self, records: list[DataRecord | Node]) -> None:
        del self.items
        for record in records:
            if not isinstance(record, DataRecord):
                DataRecord(extend=record)
            self.addChild(node=record)

    @items.deleter
    def items(self) -> None:
        for record in self.getTags("item"):
            self.delChild(record)

    def iter_records(self) -> Iterator[DataRecord | Node]:
        yield from self.getTags("item")



from test.lib.util import StanzaHandlerTest

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import LocationData
from nbxmpp.structs import PubSubEventData
from nbxmpp.structs import StanzaHandler


class LocationTest(StanzaHandlerTest):

    def test_location_parsing(self):
        def _on_message(_con, _stanza, properties):

            data = LocationData(
                accuracy="20",
                alt="1609",
                altaccuracy="10",
                area="Central Park",
                bearing="12.33",
                building="The Empire State Building",
                country="United States",
                countrycode="US",
                datum="Some datum",
                description="Bill's house",
                error="290.8882087",
                floor="102",
                lat="39.75",
                locality="New York City",
                lon="-104.99",
                postalcode="10118",
                region="New York",
                room="Observatory",
                speed="52.69",
                street="350 Fifth Avenue / 34th and Broadway",
                text="Northwest corner of the lobby",
                timestamp="2004-02-19T21:12Z",
                tzo="-07:00",
                uri="http://www.nyc.com/",
            )

            pubsub_event = PubSubEventData(
                node="http://jabber.org/protocol/geoloc",
                id="d81a52b8-0f9c-11dc-9bc8-001143d5d5db",
                item=None,
                data=data,
                deleted=False,
                retracted=False,
                purged=False,
            )

            # We cant compare Node objects
            pubsub_event_ = properties.pubsub_event._replace(item=None)
            self.assertEqual(pubsub_event, pubsub_event_)

        event = """
            <message from='test@test.test'>
                <event xmlns='http://jabber.org/protocol/pubsub#event'>
                    <items node='http://jabber.org/protocol/geoloc'>
                        <item id='d81a52b8-0f9c-11dc-9bc8-001143d5d5db'>
                            <geoloc xmlns='http://jabber.org/protocol/geoloc' xml:lang='en'>
                                <accuracy>20</accuracy>
                                <alt>1609</alt>
                                <altaccuracy>10</altaccuracy>
                                <area>Central Park</area>
                                <bearing>12.33</bearing>
                                <building>The Empire State Building</building>
                                <country>United States</country>
                                <countrycode>US</countrycode>
                                <datum>Some datum</datum>
                                <description>Bill's house</description>
                                <error>290.8882087</error>
                                <floor>102</floor>
                                <lat>39.75</lat>
                                <locality>New York City</locality>
                                <lon>-104.99</lon>
                                <postalcode>10118</postalcode>
                                <region>New York</region>
                                <room>Observatory</room>
                                <speed>52.69</speed>
                                <street>350 Fifth Avenue / 34th and Broadway</street>
                                <text>Northwest corner of the lobby</text>
                                <timestamp>2004-02-19T21:12Z</timestamp>
                                <tzo>-07:00</tzo>
                                <uri>http://www.nyc.com/</uri>
                            </geoloc>
                        </item>
                    </items>
                </event>
            </message>
        """

        self.dispatcher.register_handler(
            StanzaHandler(
                name="message", callback=_on_message, ns=Namespace.PUBSUB_EVENT
            )
        )

        self.dispatcher.process_data(event)



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import process_response
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import IqProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Ping(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="iq",
                callback=self._process_ping,
                typ="get",
                ns=Namespace.PING,
                priority=15,
            ),
        ]

    def _process_ping(
        self, _client: Client, stanza: Iq, _properties: IqProperties
    ) -> None:
        self._log.info("Send pong to %s", stanza.getFrom())
        iq = stanza.buildSimpleReply("result")
        self._client.send_stanza(iq)
        raise NodeProcessed

    @iq_request_task
    def ping(self, jid: JID):
        _task = yield

        response = yield _make_ping_request(jid)
        yield process_response(response)


def _make_ping_request(jid: JID) -> Iq:
    iq = Iq("get", to=jid)
    iq.addChild(name="ping", namespace=Namespace.PING)
    return iq



# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

# Initial version taken from https://github.com/horazont/aioxmpp
# Modified on 30-AUG-2022

from __future__ import annotations

import stringprep
from collections.abc import Callable
from unicodedata import ucd_3_2_0

_nodeprep_prohibited = frozenset("\"&'/:<>@")


def is_RandALCat(c: str) -> bool:
    return ucd_3_2_0.bidirectional(c) in ("R", "AL")


def is_LCat(c: str) -> bool:
    return ucd_3_2_0.bidirectional(c) == "L"


def check_nodeprep_prohibited(char: str) -> bool:
    return char in _nodeprep_prohibited


def _check_against_tables(
    chars: list[str], tables: tuple[Callable[[str], bool], ...]
) -> str | None:
    """
    Perform a check against the table predicates in `tables`. `tables` must be
    a reusable iterable containing characteristic functions of character sets,
    that is, functions which return :data:`True` if the character is in the
    table.
    The function returns the first character occuring in any of the tables or
    :data:`None` if no character matches.
    """

    for c in chars:
        if any(in_table(c) for in_table in tables):
            return c

    return None


def do_normalization(chars: list[str]) -> None:
    """
    Perform the stringprep normalization. Operates in-place on a list of
    unicode characters provided in `chars`.
    """
    chars[:] = list(ucd_3_2_0.normalize("NFKC", "".join(chars)))


def check_bidi(chars: list[str]) -> None:
    """
    Check proper bidirectionality as per stringprep. Operates on a list of
    unicode characters provided in `chars`.
    """

    # the empty string is valid, as it cannot violate the RandALCat constraints
    if not chars:
        return

    # first_is_RorAL = ucd_3_2_0.bidirectional(chars[0]) in {"R", "AL"}
    # if first_is_RorAL:

    has_RandALCat = any(is_RandALCat(c) for c in chars)
    if not has_RandALCat:
        return

    has_LCat = any(is_LCat(c) for c in chars)
    if has_LCat:
        raise ValueError("L and R/AL characters must not occur in the same" " string")

    if not is_RandALCat(chars[0]) or not is_RandALCat(chars[-1]):
        raise ValueError("R/AL string must start and end with R/AL character.")


def check_against_tables(
    chars: list[str], bad_tables: tuple[Callable[[str], bool], ...]
) -> None:
    """
    Check against tables, by checking whether any of the characters
    from `chars` are in any of the `bad_tables`.
    Operates in-place on a list of code points from `chars`.
    """
    violator = _check_against_tables(chars, bad_tables)
    if violator is not None:
        raise ValueError(
            "Input contains prohibited or unassigned codepoint: "
            "U+{:04x}".format(ord(violator))
        )


def _nodeprep_do_mapping(chars: list[str]) -> None:
    i = 0
    while i < len(chars):
        c = chars[i]
        if stringprep.in_table_b1(c):
            del chars[i]
        else:
            replacement = stringprep.map_table_b2(c)
            if replacement != c:
                chars[i : (i + 1)] = list(replacement)
            i += len(replacement)


def nodeprep(string: str, allow_unassigned: bool = False) -> str:
    """
    Process the given `string` using the Nodeprep (`RFC 6122`_) profile. In the
    error cases defined in `RFC 3454`_ (stringprep), a :class:`ValueError` is
    raised.
    """

    chars = list(string)
    _nodeprep_do_mapping(chars)
    do_normalization(chars)
    check_against_tables(
        chars,
        (
            stringprep.in_table_c11,
            stringprep.in_table_c12,
            stringprep.in_table_c21,
            stringprep.in_table_c22,
            stringprep.in_table_c3,
            stringprep.in_table_c4,
            stringprep.in_table_c5,
            stringprep.in_table_c6,
            stringprep.in_table_c7,
            stringprep.in_table_c8,
            stringprep.in_table_c9,
            check_nodeprep_prohibited,
        ),
    )
    check_bidi(chars)

    if not allow_unassigned:
        check_against_tables(chars, (stringprep.in_table_a1,))

    return "".join(chars)


def _resourceprep_do_mapping(chars: list[str]) -> None:
    i = 0
    while i < len(chars):
        c = chars[i]
        if stringprep.in_table_b1(c):
            del chars[i]
            continue
        i += 1


def resourceprep(string: str, allow_unassigned: bool = False) -> str:
    """
    Process the given `string` using the Resourceprep (`RFC 6122`_) profile. In
    the error cases defined in `RFC 3454`_ (stringprep), a :class:`ValueError`
    is raised.
    """

    chars = list(string)
    _resourceprep_do_mapping(chars)
    do_normalization(chars)
    check_against_tables(
        chars,
        (
            stringprep.in_table_c12,
            stringprep.in_table_c21,
            stringprep.in_table_c22,
            stringprep.in_table_c3,
            stringprep.in_table_c4,
            stringprep.in_table_c5,
            stringprep.in_table_c6,
            stringprep.in_table_c7,
            stringprep.in_table_c8,
            stringprep.in_table_c9,
        ),
    )
    check_bidi(chars)

    if not allow_unassigned:
        check_against_tables(chars, (stringprep.in_table_a1,))

    return "".join(chars)


def nameprep(string: str, allow_unassigned: bool = False) -> str:
    """
    Process the given `string` using the Nameprep (`RFC 3491`_) profile. In the
    error cases defined in `RFC 3454`_ (stringprep), a :class:`ValueError` is
    raised.
    """

    chars = list(string)
    _nodeprep_do_mapping(chars)
    do_normalization(chars)
    check_against_tables(
        chars,
        (
            stringprep.in_table_c12,
            stringprep.in_table_c22,
            stringprep.in_table_c3,
            stringprep.in_table_c4,
            stringprep.in_table_c5,
            stringprep.in_table_c6,
            stringprep.in_table_c7,
            stringprep.in_table_c8,
            stringprep.in_table_c9,
        ),
    )
    check_bidi(chars)

    if not allow_unassigned:
        check_against_tables(chars, (stringprep.in_table_a1,))

    return "".join(chars)


def _saslprep_do_mapping(chars: list[str]) -> None:
    i = 0
    while i < len(chars):
        c = chars[i]
        if stringprep.in_table_b1(c):
            del chars[i]

        elif stringprep.in_table_c12(c):
            chars[i] = " "

        i += 1


def saslprep(string: str, allow_unassigned: bool = False) -> str:
    """
    Process the given `string` using the SASLprep (`RFC 4013`_) profile.
    """

    chars = list(string)
    _saslprep_do_mapping(chars)
    do_normalization(chars)
    check_against_tables(
        chars,
        (
            stringprep.in_table_c12,
            stringprep.in_table_c21,
            stringprep.in_table_c22,
            stringprep.in_table_c3,
            stringprep.in_table_c4,
            stringprep.in_table_c5,
            stringprep.in_table_c6,
            stringprep.in_table_c7,
            stringprep.in_table_c8,
            stringprep.in_table_c9,
        ),
    )
    check_bidi(chars)

    if not allow_unassigned:
        check_against_tables(chars, (stringprep.in_table_a1,))

    return "".join(chars)



# Copyright (C) 2018-2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import typing
from typing import Any
from typing import NamedTuple
from typing import Optional
from typing import TYPE_CHECKING

import logging
import random
import time
from collections.abc import Sequence
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime

from gi.repository import Gio
from gi.repository import GLib

from nbxmpp import exceptions
from nbxmpp.const import AdHocAction
from nbxmpp.const import AdHocNoteType
from nbxmpp.const import AdHocStatus
from nbxmpp.const import Affiliation
from nbxmpp.const import AnonymityMode
from nbxmpp.const import AvatarState
from nbxmpp.const import Chatstate
from nbxmpp.const import ConnectionProtocol
from nbxmpp.const import ConnectionType
from nbxmpp.const import InviteType
from nbxmpp.const import IqType
from nbxmpp.const import MessageType
from nbxmpp.const import PresenceShow
from nbxmpp.const import PresenceType
from nbxmpp.const import Role
from nbxmpp.const import StatusCode
from nbxmpp.language import LanguageMap
from nbxmpp.language import LanguageRange
from nbxmpp.language import LanguageTag
from nbxmpp.modules.dataforms import DataForm
from nbxmpp.modules.fallback import FallbacksForT
from nbxmpp.modules.fallback import strip_fallback
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import Protocol
from nbxmpp.simplexml import Node

if TYPE_CHECKING:
    from nbxmpp.modules.security_labels import SecurityLabel

log = logging.getLogger("nbxmpp.structs")


class StanzaHandler(NamedTuple):
    name: str
    callback: Any
    typ: str = ""
    ns: str = ""
    xmlns: str | None = None
    priority: int = 50


class CommonResult(NamedTuple):
    jid: JID | None = None


class InviteData(NamedTuple):
    muc: JID
    from_: JID
    reason: str | None
    password: str | None
    type: InviteType
    continued: bool
    thread: str | None


class DeclineData(NamedTuple):
    muc: JID
    from_: JID
    reason: str | None


class CaptchaData(NamedTuple):
    form: Any
    bob_data: BobData


class BobData(NamedTuple):
    algo: str
    hash_: str
    max_age: int
    data: bytes
    cid: str
    type: str


class BookmarkData(NamedTuple):
    jid: JID
    name: str | None = None
    nick: str | None = None
    autojoin: bool = False
    password: str | None = None
    extensions: Node | None = None


class VoiceRequest(NamedTuple):
    jid: JID
    nick: str
    form: Any


class MucUserData(NamedTuple):
    jid: JID | None
    affiliation: Affiliation | None
    nick: str | None
    role: Role | None
    actor: str | None
    reason: str | None


class MucDestroyed(NamedTuple):
    alternate: JID | None
    reason: str | None
    password: str | None


class MucConfigResult(NamedTuple):
    jid: JID
    form: Optional[Any] = None


class MucSubject(NamedTuple):
    text: str
    author: str | None
    timestamp: float | None


class AffiliationResult(NamedTuple):
    jid: JID
    users: dict[JID, dict[str, str]]


class EntityCapsData(NamedTuple):
    hash: str
    node: str
    ver: str


class HTTPAuthData(NamedTuple):
    id: str | None
    method: str | None
    url: str | None
    body: str | None


class StanzaIDData(NamedTuple):
    id: str
    by: str


class PubSubEventData(NamedTuple):
    node: str
    id: str | None = None
    item: Node | None = None
    data: Optional[Any] = None
    deleted: bool = False
    retracted: bool = False
    purged: bool = False


class MoodData(NamedTuple):
    mood: str
    test: str


class BlockingPush(NamedTuple):
    block: set[JID]
    unblock: set[JID]
    unblock_all: bool


class ActivityData(NamedTuple):
    activity: str
    subactivity: str
    text: str


class LocationData(NamedTuple):
    accuracy: str | None = None
    alt: str | None = None
    altaccuracy: str | None = None
    area: str | None = None
    bearing: str | None = None
    building: str | None = None
    country: str | None = None
    countrycode: str | None = None
    datum: str | None = None
    description: str | None = None
    error: str | None = None
    floor: str | None = None
    lat: str | None = None
    locality: str | None = None
    lon: str | None = None
    postalcode: str | None = None
    region: str | None = None
    room: str | None = None
    speed: str | None = None
    street: str | None = None
    text: str | None = None
    timestamp: str | None = None
    tzo: str | None = None
    uri: str | None = None


class PGPPublicKey(NamedTuple):
    jid: JID
    key: bytes
    date: float


class PGPKeyMetadata(NamedTuple):
    jid: JID
    fingerprint: str
    date: float


class OMEMOMessage(NamedTuple):
    sid: int
    iv: bytes
    keys: dict[int, tuple[bytes, bool]]
    payload: bytes


class AnnotationNote(NamedTuple):
    jid: JID
    data: str
    cdate: datetime | None = None
    mdate: datetime | None = None


class EMEData(NamedTuple):
    name: str
    namespace: str


class MuclumbusResult(NamedTuple):
    first: str | None
    last: str | None
    max: int | None
    end: bool
    items: list[MuclumbusItem]


class MuclumbusItem(NamedTuple):
    jid: str
    name: str
    nusers: str
    description: str
    language: str
    is_open: bool
    anonymity_mode: AnonymityMode


class SoftwareVersionResult(NamedTuple):
    name: str
    version: str
    os: str | None


class AdHocCommandNote(NamedTuple):
    text: str
    type: AdHocNoteType


class IBBData(NamedTuple):
    type: str
    sid: str
    block_size: int | None = None
    seq: int | None = None
    data: bytes | None = None


class OOBData(NamedTuple):
    url: str
    desc: str


class CorrectionData(NamedTuple):
    id: str


class ReplyData(NamedTuple):
    to: JID
    id: str


class ModerationData(NamedTuple):
    stanza_id: str
    by: JID | None
    reason: str | None
    stamp: datetime
    is_tombstone: bool
    occupant_id: str | None


class RetractionData(NamedTuple):
    id: str | None
    is_tombstone: bool
    timestamp: float | None


class DiscoItems(NamedTuple):
    jid: JID
    node: str
    items: list[DiscoItem]


class DiscoItem(NamedTuple):
    jid: JID
    name: str | None
    node: str | None


class RegisterData(NamedTuple):
    instructions: str | None
    form: Optional[Any]
    fields_form: Optional[Any]
    oob_url: str | None
    bob_data: BobData | None


class HTTPUploadData(NamedTuple):
    put_uri: str
    get_uri: str
    headers: dict[str, str]


class RSMData(NamedTuple):
    after: str | None
    before: str | None
    last: str | None
    first: str | None
    first_index: int | None
    count: int | None
    max: int | None
    index: int | None


class MAMQueryData(NamedTuple):
    jid: JID
    rsm: RSMData
    complete: bool


class MAMPreferencesData(NamedTuple):
    default: str
    always: list[JID]
    never: list[JID]


class LastActivityData(NamedTuple):
    seconds: int
    status: str


class RosterData(NamedTuple):
    items: list[RosterItem] | None
    version: str | None


class RosterPush(NamedTuple):
    item: RosterItem
    version: str


class ServerAddress(NamedTuple):
    domain: str | None
    service: str | None
    host: str | None
    uri: str | None
    protocol: ConnectionProtocol
    type: ConnectionType
    proxy: ProxyData | None

    @property
    def is_service(self) -> bool:
        return self.service is not None

    @property
    def is_host(self) -> bool:
        return self.host is not None

    @property
    def is_uri(self) -> bool:
        return self.uri is not None

    def has_proxy(self) -> bool:
        return self.proxy is not None


@dataclass
class RosterItem:
    jid: JID
    name: str | None = None
    ask: str | None = None
    subscription: str | None = None
    approved: str | None = None
    groups: set[str] = field(default_factory=set)

    @classmethod
    def from_node(cls, node: Node) -> RosterItem:
        attrs = node.getAttrs(copy=True)
        jid = attrs.get("jid")
        if jid is None:
            raise Exception("jid attribute missing")

        jid = JID.from_string(jid)
        if jid.is_full:
            raise Exception("full jid in roster not allowed")

        groups = {group.getData() for group in node.getTags("group")}

        return cls(
            jid=jid,
            name=attrs.get("name"),
            ask=attrs.get("ask"),
            subscription=attrs.get("subscription") or "none",
            approved=attrs.get("approved"),
            groups=groups,
        )

    def asdict(self) -> dict[str, Any]:
        return {
            "jid": self.jid,
            "name": self.name,
            "ask": self.ask,
            "subscription": self.subscription,
            "approved": self.approved,
            "groups": self.groups,
        }


class DiscoInfo(NamedTuple):
    stanza: Iq | None
    identities: list[DiscoIdentity]
    features: list[str]
    dataforms: list[DataForm]
    timestamp: float | None = None

    def get_caps_hash(self) -> str | None:
        try:
            return self.node.split("#")[1]
        except Exception:
            return None

    def has_field(self, form_type: str, var: str) -> bool:
        for dataform in self.dataforms:
            try:
                if dataform["FORM_TYPE"].value != form_type:
                    continue
                if var in dataform.vars:
                    return True

            except Exception:
                continue
        return False

    def get_field_value(self, form_type: str, var: str) -> Optional[Any]:
        for dataform in self.dataforms:
            try:
                if dataform["FORM_TYPE"].value != form_type:
                    continue

                if dataform[var].type_ == "jid-multi":
                    return dataform[var].values or None
                return dataform[var].value or None

            except Exception:
                continue

        return None

    def supports(self, feature: str) -> bool:
        return feature in self.features

    def serialize(self) -> str:
        if self.stanza is None:
            raise ValueError("Unable to serialize DiscoInfo, no stanza found")
        return str(self.stanza)

    @property
    def node(self) -> str | None:
        try:
            query = self.stanza.getQuery()
        except Exception:
            return None

        if query is not None:
            return query.getAttr("node")
        return None

    @property
    def jid(self) -> JID | None:
        try:
            return self.stanza.getFrom()
        except Exception:
            return None

    @property
    def mam_namespace(self) -> str | None:
        if Namespace.MAM_2 in self.features:
            return Namespace.MAM_2
        if Namespace.MAM_1 in self.features:
            return Namespace.MAM_1
        return None

    @property
    def has_mam_2(self) -> bool:
        return Namespace.MAM_2 in self.features

    @property
    def has_mam_1(self) -> bool:
        return Namespace.MAM_1 in self.features

    @property
    def has_mam(self) -> bool:
        return self.has_mam_1 or self.has_mam_2

    @property
    def has_httpupload(self) -> bool:
        return Namespace.HTTPUPLOAD_0 in self.features

    @property
    def has_message_moderation(self) -> bool:
        return (
            Namespace.MESSAGE_MODERATE in self.features
            or Namespace.MESSAGE_MODERATE_1 in self.features
        )

    @property
    def moderation_namespace(self) -> str | None:
        if Namespace.MESSAGE_MODERATE_1 in self.features:
            return Namespace.MESSAGE_MODERATE_1
        if Namespace.MESSAGE_MODERATE in self.features:
            return Namespace.MESSAGE_MODERATE
        return None

    @property
    def is_muc(self) -> bool:
        for identity in self.identities:
            if identity.category == "conference":
                if Namespace.MUC in self.features:
                    return True
        return False

    @property
    def is_irc(self) -> bool:
        for identity in self.identities:
            if identity.category == "conference" and identity.type == "irc":
                return True
        return False

    @property
    def muc_name(self) -> str | None:
        if self.muc_room_name:
            return self.muc_room_name

        if self.muc_identity_name:
            return self.muc_identity_name

        if self.jid is not None:
            return self.jid.localpart
        return None

    @property
    def muc_identity_name(self) -> str | None:
        for identity in self.identities:
            if identity.category == "conference":
                return identity.name
        return None

    @property
    def muc_room_name(self) -> Optional[Any]:
        return self.get_field_value(Namespace.MUC_INFO, "muc#roomconfig_roomname")

    @property
    def muc_description(self) -> Optional[Any]:
        return self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_description")

    @property
    def muc_log_uri(self) -> Optional[Any]:
        return self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_logs")

    @property
    def muc_users(self) -> Optional[Any]:
        return self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_occupants")

    @property
    def muc_contacts(self) -> Optional[Any]:
        return self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_contactjid")

    @property
    def muc_subject(self) -> Optional[Any]:
        return self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_subject")

    @property
    def muc_subjectmod(self) -> Optional[Any]:
        # muc#roominfo_changesubject stems from a wrong example in the MUC XEP
        # Ejabberd and Prosody use this value
        # muc#roomconfig_changesubject is also used by Prosody
        return (
            self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_subjectmod")
            or self.get_field_value(Namespace.MUC_INFO, "muc#roomconfig_changesubject")
            or self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_changesubject")
        )

    @property
    def muc_lang(self) -> Optional[Any]:
        return self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_lang")

    @property
    def muc_is_persistent(self) -> bool:
        return "muc_persistent" in self.features

    @property
    def muc_is_moderated(self) -> bool:
        return "muc_moderated" in self.features

    @property
    def muc_is_open(self) -> bool:
        return "muc_open" in self.features

    @property
    def muc_is_members_only(self) -> bool:
        return "muc_membersonly" in self.features

    @property
    def muc_is_hidden(self) -> bool:
        return "muc_hidden" in self.features

    @property
    def muc_is_nonanonymous(self) -> bool:
        return "muc_nonanonymous" in self.features

    @property
    def muc_is_passwordprotected(self) -> bool:
        return "muc_passwordprotected" in self.features

    @property
    def muc_is_public(self) -> bool:
        return "muc_public" in self.features

    @property
    def muc_is_semianonymous(self) -> bool:
        return "muc_semianonymous" in self.features

    @property
    def muc_is_temporary(self) -> bool:
        return "muc_temporary" in self.features

    @property
    def muc_is_unmoderated(self) -> bool:
        return "muc_unmoderated" in self.features

    @property
    def muc_is_unsecured(self) -> bool:
        return "muc_unsecured" in self.features

    @property
    def is_gateway(self) -> bool:
        return any(identity.category == "gateway" for identity in self.identities)

    @property
    def gateway_name(self) -> str | None:
        for identity in self.identities:
            if identity.category == "gateway":
                return identity.name
        return None

    @property
    def gateway_type(self) -> str | None:
        for identity in self.identities:
            if identity.category == "gateway":
                return identity.type
        return None

    def has_category(self, category: str) -> bool:
        return any(identity.category == category for identity in self.identities)

    def has_identity(self, category: str, type_: str) -> bool:
        for identity in self.identities:
            if identity.category == category and identity.type == type_:
                return True
        return False

    @property
    def httpupload_max_file_size(self) -> float | None:
        size = self.get_field_value(Namespace.HTTPUPLOAD_0, "max-file-size")
        try:
            return float(size)
        except Exception:
            return None


class DiscoIdentity(NamedTuple):

    category: str
    type: str
    name: str | None = None
    lang: str | None = None

    def get_node(self) -> Node:
        identity = Node(
            "identity", attrs={"category": self.category, "type": self.type}
        )
        if self.name is not None:
            identity.setAttr("name", self.name)

        if self.lang is not None:
            identity.setAttr("xml:lang", self.lang)
        return identity

    def __eq__(self, other: object) -> bool:
        return str(self) == str(other)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return "%s/%s/%s/%s" % (
            self.category,
            self.type,
            self.lang or "",
            self.name or "",
        )

    def __hash__(self) -> int:
        return hash(str(self))


class AdHocCommand(NamedTuple):
    jid: JID
    node: Node
    name: str | None
    sessionid: str | None = None
    status: AdHocStatus | None = None
    data: Node | None = None
    actions: set[AdHocAction] | None = None
    default: AdHocAction | None = None
    notes: list[AdHocCommandNote] | None = None

    @property
    def is_completed(self) -> bool:
        return self.status == AdHocStatus.COMPLETED

    @property
    def is_canceled(self) -> bool:
        return self.status == AdHocStatus.CANCELED


class ProxyData(NamedTuple):
    type: str
    host: str
    username: str | None
    password: str | None

    def get_uri(self) -> str:
        if self.username is not None:
            username = GLib.uri_escape_string(self.username, None, False)
            password = GLib.uri_escape_string(self.password, None, False)
            user_pass = f"{username}:{password}"
            return "%s://%s@%s" % (self.type, user_pass, self.host)
        return "%s://%s" % (self.type, self.host)

    def get_resolver(self) -> Gio.SimpleProxyResolver:
        return Gio.SimpleProxyResolver.new(self.get_uri(), None)


class OMEMOBundle(NamedTuple):
    spk: dict[str, int | bytes]
    spk_signature: bytes
    ik: bytes
    otpks: list[dict[str, str]]
    device_id: int = -1
    namespace: str = Namespace.OMEMO_TEMP

    def pick_prekey(self) -> dict[str, str]:
        return random.SystemRandom().choice(self.otpks)


class ChatMarker(NamedTuple):
    type: str
    id: str

    @property
    def is_received(self) -> bool:
        return self.type == "received"

    @property
    def is_displayed(self) -> bool:
        return self.type == "displayed"

    @property
    def is_acknowledged(self) -> bool:
        return self.type == "acknowledged"


class Reactions(NamedTuple):
    id: str
    emojis: set[str]


class CommonError:
    def __init__(self, stanza: Protocol) -> None:
        self._stanza_name = stanza.getName()
        self._error_node = stanza.getTag("error")
        self.condition = stanza.getError()
        self.condition_data = self._error_node.getTagData(self.condition)
        self.app_condition = stanza.getAppError()
        self.type = stanza.getErrorType()
        self.by = None
        self.jid = stanza.getFrom()
        self.id = stanza.getID()
        self._text: dict[str, str] = {}

        by = self._error_node.getAttr("by")
        if by is not None:
            try:
                self.by = JID.from_string(by)
            except Exception:
                pass

        text_elements = self._error_node.getTags("text", namespace=Namespace.STANZAS)
        for element in text_elements:
            lang = element.getXmlLang()
            text = element.getData()
            self._text[lang] = text

    @classmethod
    def from_string(cls, node_string: bytes | str) -> CommonError:
        return cls(Protocol(node=node_string))

    def get_text(self, pref_lang: str | None = None) -> str:
        if pref_lang is not None:
            text = self._text.get(pref_lang)
            if text is not None:
                return text

        if self._text:
            text = self._text.get("en")
            if text is not None:
                return text

            text = self._text.get(None)
            if text is not None:
                return text
            return self._text.popitem()[1]
        return ""

    def set_text(self, lang: str, text: str) -> None:
        self._text[lang] = text

    def __str__(self) -> str:
        condition = self.condition
        if self.app_condition is not None:
            condition = "%s (%s)" % (self.condition, self.app_condition)
        text = self.get_text("en") or ""
        if text:
            text = " - %s" % text
        return "Error from %s: %s%s" % (self.jid, condition, text)

    def serialize(self) -> str:
        return str(
            Protocol(
                name=self._stanza_name,
                frm=self.jid,
                xmlns=Namespace.CLIENT,
                attrs={"id": self.id},
                payload=self._error_node,
            )
        )


class HTTPUploadError(CommonError):
    def __init__(self, stanza: Protocol) -> None:
        CommonError.__init__(self, stanza)

    def get_max_file_size(self) -> float | None:
        if not self.app_condition == "file-too-large":
            return None
        node = self._error_node.getTag(self.app_condition)
        try:
            return float(node.getTagData("max-file-size"))
        except Exception:
            return None

    def get_retry_date(self):
        if not self.app_condition == "retry":
            return None
        return self._error_node.getTagAttr("stamp")


class StanzaMalformedError(CommonError):
    def __init__(self, stanza: Protocol, text: str | None) -> None:
        self._error_node = None
        self.condition = "stanza-malformed"
        self.condition_data = None
        self.app_condition = None
        self.type = None
        self.jid = stanza.getFrom()
        self.id = stanza.getID()
        self._text: dict[str, str] = {}
        if text:
            self._text["en"] = text

    @classmethod
    def from_string(cls, node_string: str) -> Any:
        raise NotImplementedError

    def __str__(self) -> str:
        text = self.get_text("en")
        if text:
            text = ": %s" % text
        return "Received malformed stanza from %s%s" % (self.jid, text)

    def serialize(self) -> str:
        raise NotImplementedError


class StreamError(CommonError):
    def __init__(self, stanza: Protocol) -> None:
        self.condition = stanza.getError()
        self.condition_data = self._error_node.getTagData(self.condition)
        self.app_condition = stanza.getAppError()
        self.type = stanza.getErrorType()
        self.jid = stanza.getFrom()
        self.id = stanza.getID()
        self._text: dict[str, str] = {}

        text_elements = self._error_node.getTags("text", namespace=Namespace.STREAMS)
        for element in text_elements:
            lang = element.getXmlLang()
            text = element.getData()
            self._text[lang] = text

    @classmethod
    def from_string(cls, node_string: str) -> Any:
        raise NotImplementedError

    def __str__(self) -> str:
        text = self.get_text("en") or ""
        if text:
            text = " - %s" % text
        return "Error from %s: %s%s" % (self.jid, self.condition, text)

    def serialize(self) -> str:
        raise NotImplementedError


class TuneData(NamedTuple):
    artist: str | None = None
    length: str | None = None
    rating: str | None = None
    source: str | None = None
    title: str | None = None
    track: str | None = None
    uri: str | None = None

    @property
    def was_removed(self) -> bool:
        return self.artist is None and self.title is None and self.track is None


class MAMData(NamedTuple):
    id: str
    query_id: str
    archive: JID
    namespace: str
    timestamp: float

    @property
    def is_ver_1(self) -> bool:
        return self.namespace == Namespace.MAM_1

    @property
    def is_ver_2(self) -> bool:
        return self.namespace == Namespace.MAM_2


class CarbonData(NamedTuple):
    type: str

    @property
    def is_sent(self) -> bool:
        return self.type == "sent"

    @property
    def is_received(self) -> bool:
        return self.type == "received"


class ReceiptData(NamedTuple):
    type: str
    id: str | None = None

    @property
    def is_request(self) -> bool:
        return self.type == "request"

    @property
    def is_received(self) -> bool:
        return self.type == "received"


@dataclass
class Hat:
    uri: str
    title: str


class HatData:
    def __init__(self) -> None:
        self._hat_map = LanguageMap()

    def add_hat(self, hat: Hat, lang: str | None) -> None:
        language_tag = LanguageTag(tag=lang) if lang else None
        if language_tag not in self._hat_map:
            self._hat_map[language_tag] = []
        self._hat_map[language_tag].append(hat)

    def get_hats(
        self, language_range: Sequence[LanguageRange] | None = None
    ) -> list[Hat]:

        if language_range is None:
            return self._hat_map.any()

        try:
            return self._hat_map.lookup(language_range)
        except KeyError:
            return self._hat_map.any()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HatData):
            return False
        return self._hat_map == other._hat_map

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)


@dataclass
class EncryptionData:
    protocol: str
    key: str | None
    trust: int


class Properties:
    pass


@dataclass
class MessageProperties:
    own_jid: JID
    carbon: CarbonData | None = None
    type: MessageType = MessageType.NORMAL
    id: str | None = None
    stanza_ids: list[StanzaIDData] = field(default_factory=list)
    origin_id: str | None = None
    from_: JID | None = None
    to: JID | None = None
    jid: JID | None = None
    remote_jid: JID | None = None
    subject: str | None = None
    body: str | None = None
    bodies: BodyData | None = None
    thread: str | None = None
    user_timestamp: float | None = None
    timestamp: float = field(default_factory=time.time)
    has_server_delay: bool = False
    error = None
    eme: EMEData | None = None
    http_auth: HTTPAuthData | None = None
    nickname: str | None = None
    from_muc: bool = False
    occupant_id: str | None = None
    muc_jid: JID | None = None
    muc_nickname: str | None = None
    muc_status_codes: set[StatusCode] | None = None
    muc_private_message: bool = False
    muc_invite: InviteData | None = None
    muc_decline: DeclineData | None = None
    muc_user: MucUserData | None = None
    muc_ofrom: JID | None = None
    muc_subject: MucSubject | None = None
    captcha: CaptchaData | None = None
    voice_request: VoiceRequest | None = None
    self_message: bool = False
    mam: MAMData | None = None
    pubsub: bool = False
    pubsub_event: PubSubEventData | None = None
    openpgp: bytes | None = None
    omemo: OMEMOMessage | None = None
    encrypted: EncryptionData | None = None
    pgp_legacy: str | None = None
    marker: ChatMarker | None = None
    receipt: ReceiptData | None = None
    oob: OOBData | None = None
    correction: CorrectionData | None = None
    reply_data: ReplyData | None = None
    moderation: ModerationData | None = None
    retraction: RetractionData | None = None
    attention: bool = False
    forms = None
    xhtml: XHTMLData | None = None
    security_label: SecurityLabel | None = None
    chatstate: Chatstate | None = None
    reactions: Reactions | None = None

    def is_from_us(self, bare_match: bool = True) -> bool:
        if self.from_ is None:
            raise ValueError("from attribute missing")

        if bare_match:
            return self.own_jid.bare_match(self.from_)
        return self.own_jid == self.from_

    @property
    def has_user_delay(self) -> bool:
        return self.user_timestamp is not None

    @property
    def is_encrypted(self) -> bool:
        return self.encrypted is not None

    @property
    def is_omemo(self) -> bool:
        return self.omemo is not None

    @property
    def is_openpgp(self) -> bool:
        return self.openpgp is not None

    @property
    def is_pgp_legacy(self) -> bool:
        return self.pgp_legacy is not None

    @property
    def is_pubsub(self) -> bool:
        return self.pubsub

    @property
    def is_pubsub_event(self) -> bool:
        return self.pubsub_event is not None

    @property
    def is_carbon_message(self) -> bool:
        return self.carbon is not None

    @property
    def is_sent_carbon(self) -> bool:
        return self.carbon is not None and self.carbon.is_sent

    @property
    def is_received_carbon(self) -> bool:
        return self.carbon is not None and self.carbon.is_received

    @property
    def is_mam_message(self) -> bool:
        return self.mam is not None

    @property
    def is_http_auth(self) -> bool:
        return self.http_auth is not None

    @property
    def is_muc_subject(self) -> bool:
        return (
            self.type == MessageType.GROUPCHAT
            and self.body is None
            and self.subject is not None
        )

    @property
    def is_muc_config_change(self) -> bool:
        return bool(self.muc_status_codes)

    @property
    def is_muc_pm(self) -> bool:
        return self.muc_private_message

    @property
    def is_muc_invite_or_decline(self) -> bool:
        return self.muc_invite is not None or self.muc_decline is not None

    @property
    def is_captcha_challenge(self) -> bool:
        return self.captcha is not None

    @property
    def is_voice_request(self) -> bool:
        return self.voice_request is not None

    @property
    def is_self_message(self) -> bool:
        return self.self_message

    @property
    def is_marker(self) -> bool:
        return self.marker is not None

    @property
    def is_receipt(self) -> bool:
        return self.receipt is not None

    @property
    def is_oob(self) -> bool:
        return self.oob is not None

    @property
    def is_correction(self) -> bool:
        return self.correction is not None

    @property
    def is_moderation(self) -> bool:
        return self.moderation is not None

    @property
    def is_retraction(self) -> bool:
        return self.retraction is not None

    @property
    def has_attention(self) -> bool:
        return self.attention

    @property
    def has_forms(self) -> bool:
        return self.forms is not None

    @property
    def has_xhtml(self) -> bool:
        return self.xhtml is not None

    @property
    def has_security_label(self) -> bool:
        return self.security_label is not None

    @property
    def has_chatstate(self) -> bool:
        return self.chatstate is not None


@dataclass
class IqProperties:
    own_jid: JID
    type: IqType | None = None
    jid: JID | None = None
    id: str | None = None
    error: Optional[Any] = None
    query: Node | None = None
    payload: Node | None = None
    http_auth: HTTPAuthData | None = None
    ibb: IBBData | None = None
    blocking: BlockingPush | None = None
    roster: RosterPush | None = None

    @property
    def is_http_auth(self) -> bool:
        return self.http_auth is not None

    @property
    def is_ibb(self) -> bool:
        return self.ibb is not None

    @property
    def is_blocking(self) -> bool:
        return self.blocking is not None

    @property
    def is_roster(self) -> bool:
        return self.roster is not None


class IqPropertiesBase(typing.Protocol):
    own_jid: JID
    type: IqType
    jid: JID
    id: str
    error: Optional[Any]
    query: Node | None
    payload: Node | None


class BlockingProperties(IqPropertiesBase):

    blocking: BlockingPush

    @property
    def is_blocking(self) -> bool: ...


@dataclass
class PresenceProperties:
    own_jid: JID
    type: PresenceType | None = None
    priority: int | None = None
    show: PresenceShow | None = None
    jid: JID | None = None
    resource: str | None = None
    id: str | None = None
    nickname: str | None = None
    self_presence: bool = False
    self_bare: bool = False
    from_muc: bool = False
    occupant_id: str | None = None
    status: str = ""
    timestamp: float = field(default_factory=time.time)
    user_timestamp: float | None = None
    idle_timestamp: float | None = None
    signed: Optional[Any] = None
    error: Optional[Any] = None
    avatar_sha: str | None = None
    avatar_state: AvatarState = AvatarState.IGNORE
    muc_jid: JID | None = None
    muc_status_codes: set[StatusCode] | None = None
    muc_user: MucUserData | None = None
    muc_nickname: str | None = None
    muc_destroyed: MucDestroyed | None = None
    entity_caps: EntityCapsData | None = None
    hats: HatData | None = None

    @property
    def is_self_presence(self) -> bool:
        return self.self_presence

    @property
    def is_self_bare(self) -> bool:
        return self.self_bare

    @property
    def is_muc_destroyed(self) -> bool:
        return self.muc_destroyed is not None

    @property
    def is_muc_self_presence(self) -> bool:
        return (
            self.from_muc
            and self.muc_status_codes is not None
            and StatusCode.SELF in self.muc_status_codes
        )

    @property
    def is_nickname_modified(self) -> bool:
        return (
            self.from_muc
            and self.muc_status_codes is not None
            and StatusCode.NICKNAME_MODIFIED in self.muc_status_codes
            and self.type == PresenceType.AVAILABLE
        )

    @property
    def is_nickname_changed(self) -> bool:
        return (
            self.from_muc
            and self.muc_status_codes is not None
            and StatusCode.NICKNAME_CHANGE in self.muc_status_codes
            and self.muc_user.nick is not None
            and self.type == PresenceType.UNAVAILABLE
        )

    @property
    def new_jid(self) -> JID:
        if not self.is_nickname_changed:
            raise ValueError("This is not a nickname change")
        return self.jid.new_with(resource=self.muc_user.nick)

    @property
    def is_kicked(self) -> bool:
        status_codes = {
            StatusCode.REMOVED_BANNED,
            StatusCode.REMOVED_KICKED,
            StatusCode.REMOVED_AFFILIATION_CHANGE,
            StatusCode.REMOVED_NONMEMBER_IN_MEMBERS_ONLY,
            StatusCode.REMOVED_SERVICE_SHUTDOWN,
            StatusCode.REMOVED_ERROR,
        }
        return (
            self.from_muc
            and self.muc_status_codes is not None
            and bool(status_codes.intersection(self.muc_status_codes))
            and self.type == PresenceType.UNAVAILABLE
        )

    @property
    def is_muc_shutdown(self) -> bool:
        return (
            self.from_muc
            and self.muc_status_codes is not None
            and StatusCode.REMOVED_SERVICE_SHUTDOWN in self.muc_status_codes
        )

    @property
    def is_new_room(self) -> bool:
        status_codes = {StatusCode.CREATED, StatusCode.SELF}
        return (
            self.from_muc
            and self.muc_status_codes is not None
            and status_codes.issubset(self.muc_status_codes)
        )

    @property
    def affiliation(self) -> Affiliation | None:
        try:
            return self.muc_user.affiliation
        except Exception:
            return None

    @property
    def role(self) -> Role | None:
        try:
            return self.muc_user.role
        except Exception:
            return None


class XHTMLData:
    def __init__(self, xhtml: Node) -> None:
        self._bodys: dict[str | None, Node] = {}
        for body in xhtml.getTags("body", namespace=Namespace.XHTML):
            lang = body.getXmlLang()
            self._bodys[lang] = body

    def get_body(self, pref_lang: str | None = None) -> str:
        if pref_lang is not None:
            body = self._bodys.get(pref_lang)
            if body is not None:
                return str(body)

        body = self._bodys.get("en")
        if body is not None:
            return str(body)

        body = self._bodys.get(None)
        if body is not None:
            return str(body)
        return str(self._bodys.popitem()[1])


@dataclass
class ChannelBindingData:
    type: Gio.TlsChannelBindingType
    data: bytes


@dataclass
class MDSData:
    jid: JID
    stanza_id: str
    stanza_id_by: JID


@dataclass
class BodyData:
    def __init__(
        self,
        stanza: Node,
        fallbacks_for: FallbacksForT | None = None,
        fallback_ns: set[str] | None = None,
    ) -> None:

        self._body_map = LanguageMap()
        self._fallbacks_for = fallbacks_for
        self._fallback_ns = fallback_ns

        for body in stanza.getTags("body"):
            lang = body.getAttr("xml:lang")
            lang_tag = LanguageTag(tag=lang) if lang else None
            self._body_map[lang_tag] = (lang, body.getData())

    def get(
        self,
        language_range: Sequence[LanguageRange] | None,
    ) -> str:

        if not self._body_map:
            return ""

        if language_range is None:
            lang, text = self._body_map.any()

        else:
            try:
                lang, text = self._body_map.lookup(language_range)
            except KeyError:
                lang, text = self._body_map.any()

        if not self._fallbacks_for or not self._fallback_ns:
            return text

        try:
            return strip_fallback(self._fallbacks_for, self._fallback_ns, lang, text)
        except exceptions.FallbackLanguageError:
            log.warning("Missing fallback for language: %s", lang)
            return text



from precis_i18n import get_profile

_localpart_disallowed_chars = set("\"&'/:<>@")


def enforce_precis_username(localpart: str) -> str:
    if _localpart_disallowed_chars & set(localpart):
        raise ValueError("Input contains prohibited codepoint: %s" % localpart)

    username = get_profile("UsernameCaseMapped")
    return username.enforce(localpart)


def enforce_precis_opaque(resourcepart: str) -> str:
    opaque = get_profile("OpaqueString")
    return opaque.enforce(resourcepart)



from __future__ import annotations

from typing import Any

import os
import unittest

from gi.repository import GLib

from nbxmpp.client import Client
from nbxmpp.http import HTTPSession
from nbxmpp.modules.muclumbus import Muclumbus
from nbxmpp.structs import MuclumbusResult

# Test vector from https://wiki.xmpp.org/web/SASL_and_SCRAM-SHA-1

API_URL = "https://search.jabber.network/api/1.0/search"


@unittest.skipUnless(
    os.environ.get("NBXMPP_EXTERNAL_UNIT_TESTS"), "ENV var for external tests not set"
)
class TestMuclumbus(unittest.TestCase):
    def setUp(self):
        self._client = Client()
        self._client.set_http_session(HTTPSession())
        self._module = Muclumbus(self._client)

    def test_http_request(self):

        mainloop = GLib.MainLoop()

        def _result(task: Any) -> None:
            result = task.finish()
            assert isinstance(result, MuclumbusResult)
            assert len(result.items) > 0
            mainloop.quit()

        self._module.set_http_search(API_URL, "gajim", callback=_result)

        mainloop.run()


if __name__ == "__main__":
    unittest.main()



from test.lib.util import StanzaHandlerTest

from nbxmpp import Presence
from nbxmpp.language import LanguageRange
from nbxmpp.modules.muc.hats import Hats
from nbxmpp.structs import Hat
from nbxmpp.structs import PresenceProperties


class TestHats(StanzaHandlerTest):
    def test_parse_hats(self):
        # language=XML
        xml = """
        <presence
            from='meeting123@meetings.example.com/Harry'
            id='D568A74F-E062-407C-83E9-531E91526516'
            to='someone@example.com/foo'>
          <x xmlns='http://jabber.org/protocol/muc#user'>
            <item affiliation='owner' role='moderator'/>
          </x>
          <hats xmlns='urn:xmpp:hats:0'>
            <hat title='Member' uri='http://schemas.example.com/hats#member' xml:lang='en-us' />
            <hat title='Mitglied' uri='http://schemas.example.com/hats#member' xml:lang='de-de' />
            <hat title='Owner' uri='http://schemas.example.com/hats#owner' xml:lang='en-us' />
            <hat title='Eigentümer' uri='http://schemas.example.com/hats#owner' xml:lang='de-de' />
          </hats>
        </presence>
        """

        props = PresenceProperties("mockjid")
        presence = Presence(node=xml)
        Hats._process_hats(None, None, presence, props)

        hat_list_en = [
            Hat(uri="http://schemas.example.com/hats#member", title="Member"),
            Hat(uri="http://schemas.example.com/hats#owner", title="Owner"),
        ]

        hat_list_de = [
            Hat(uri="http://schemas.example.com/hats#member", title="Mitglied"),
            Hat(uri="http://schemas.example.com/hats#owner", title="Eigentümer"),
        ]

        lang_range_en = [LanguageRange(tag="en")]
        lang_range_de = [LanguageRange(tag="de")]

        assert props.hats is not None

        hats_en = props.hats.get_hats(lang_range_en)
        self.assertSequenceEqual(hat_list_en, hats_en)

        hats_de = props.hats.get_hats(lang_range_de)
        self.assertSequenceEqual(hat_list_de, hats_de)

        hats_any = props.hats.get_hats()
        self.assertSequenceEqual(hat_list_de, hats_any)
